from typing import Literal, Optional

from pydantic import BaseModel, ConfigDict, Field

from emp_agents.models.shared.message import Message
from emp_agents.models.shared.tools import GenericTool


class Request(BaseModel):
    """
    https://platform.openai.com/docs/api-reference/chat/create
    """

    model_config = ConfigDict(populate_by_name=True)

    model: str
    max_tokens: Optional[int] = Field(default=4096, lt=128_000, gt=0)
    temperature: Optional[float] = Field(default=None, ge=0, le=2.0)
    tool_choice: Literal["none", "required", "auto", None] = Field(default=None)
    tools: Optional[list[GenericTool]] = None
    response_format: type[BaseModel] | None = None

    system: str | None = None  # anthropic field
    messages: list[Message]

    frequency_penalty: Optional[float] = Field(default=None, ge=-2.0, le=2.0)  # openai
    presence_penalty: Optional[float] = Field(default=None, ge=-2.0, le=2.0)  # openai
    num_responses: Optional[int] = Field(
        default=None, serialization_alias="n"
    )  # openai
    top_p: Optional[int] = Field(default=None)  # openai

    def model_dump(self, *, exclude_none=True, by_alias=True, **kwargs):
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )



import enum
import inspect
from inspect import isclass
from typing import (
    Annotated,
    Any,
    Callable,
    Literal,
    Optional,
    Union,
    get_args,
    get_origin,
)

from fast_depends.dependencies.model import Depends
from pydantic import BaseModel

from emp_agents.implicits.models import IgnoreDepends

from .types import Doc, FunctionSchema
from .utils import unwrap_doc

try:
    from types import UnionType
except ImportError:
    UnionType = Union  # type: ignore


__all__ = ("get_function_schema", "guess_type", "Doc", "Annotated")


def get_function_schema(  # noqa: C901
    func: Annotated[Callable, Doc("The function to get the schema for")],
    format: Annotated[
        Optional[Literal["openai", "claude"]],
        Doc("The format of the schema to return"),
    ] = "openai",
    ignore_class_arg: bool = True,
) -> Annotated[FunctionSchema, Doc("The JSON schema for the given function")]:
    """
    Returns a JSON schema for the given function.

    You can annotate your function parameters with the special Annotated type.
    Then get the schema for the function without writing the schema by hand.

    Especially useful for OpenAI API function-call.

    Example:
    >>> from typing import Annotated, Optional
    >>> import enum
    >>> def get_weather(
    ...     city: Annotated[str, Doc("The city to get the weather for")],
    ...     unit: Annotated[
    ...         Optional[str],
    ...         Doc("The unit to return the temperature in"),
    ...         enum.Enum("Unit", "celcius fahrenheit")
    ...     ] = "celcius",
    ... ) -> str:
    ...     \"\"\"Returns the weather for the given city.\"\"\"
    ...     return f"Hello {name}, you are {age} years old."
    >>> get_function_schema(get_weather) # doctest: +SKIP
    {
        'name': 'get_weather',
        'description': 'Returns the weather for the given city.',
        'parameters': {
            'type': 'object',
            'properties': {
                'city': {
                    'type': 'string',
                    'description': 'The city to get the weather for'
                },
                'unit': {
                    'type': 'string',
                    'description': 'The unit to return the temperature in',
                    'enum': ['celcius', 'fahrenheit'],
                    'default': 'celcius'
                }
            },
            'required': ['city']
        }
    }
    """
    if isinstance(func, classmethod):
        func = func.__func__
    elif isinstance(func, staticmethod):
        func = func.__func__

    sig = inspect.signature(func)
    params = sig.parameters
    schema: dict[str, Any] = {
        "type": "object",
        "properties": {},
        "required": [],
    }
    for name, param in params.items():
        param_args = get_args(param.annotation)
        is_annotated = get_origin(param.annotation) is Annotated

        enum_ = None
        default_value = inspect._empty

        if is_annotated:
            # first arg is type
            (T, *_) = param_args

            # find description in param_args tuple
            try:
                description = next(
                    unwrap_doc(arg) for arg in param_args if isinstance(arg, Doc)
                )
            except StopIteration:
                try:
                    description = next(
                        arg for arg in param_args if isinstance(arg, str)
                    )
                except StopIteration:
                    description = "The {name} parameter"

            # find enum in param_args tuple
            enum_ = next(
                (
                    [e.name for e in arg]
                    for arg in param_args
                    if isinstance(arg, type) and issubclass(arg, enum.Enum)
                ),
                # use typing.Literal as enum if no enum found
                get_origin(T) is Literal and get_args(T) or None,
            )
        else:
            T = param.annotation
            description = f"The {name} parameter"
            if get_origin(T) is Literal:
                enum_ = get_args(T)

        # find default value for param
        if param.default is not inspect._empty:
            default_value = param.default

        if default_value is not inspect._empty and isinstance(
            default_value, IgnoreDepends
        ):
            continue

        schema["properties"][name] = {
            "type": guess_type(T),
            "description": description,
        }
        if isclass(T) and issubclass(T, BaseModel):
            schema["properties"][name]["properties"] = T.model_json_schema()[
                "properties"
            ]
            schema["properties"][name]["type"] = "object"
            schema["properties"][name]["required"] = T.model_json_schema()["required"]

            for item in schema["properties"][name]["properties"]:
                del schema["properties"][name]["properties"][item]["title"]
                schema["properties"][name]["properties"][item][
                    "description"
                ] = "a field in the model"

        if guess_type(T) == "array":
            schema["properties"][name]["items"] = {"type": guess_type(get_args(T)[0])}

        if enum_ is not None:
            schema["properties"][name]["enum"] = [t for t in enum_ if t is not None]

        # TODO: check if default value is set via the depends call
        if default_value is not inspect._empty and not isinstance(
            default_value,
            Depends,
        ):
            schema["properties"][name]["default"] = default_value

        try:
            if (
                get_origin(T) is not Literal
                and not isinstance(None, T)
                and default_value is inspect._empty
            ):
                schema["required"].append(name)
        except Exception:
            schema["required"].append(name)

        if get_origin(T) is Literal:
            if all(get_args(T)):
                schema["required"].append(name)

    params_key = "input_schema" if format == "claude" else "parameters"

    schema["required"] = list(set(schema["required"]))

    return {
        "name": func.__name__,
        "description": inspect.getdoc(func),
        params_key: schema,  # type: ignore
    }


def guess_type(  # noqa: C901
    T: Annotated[type, Doc("The type to guess the JSON schema type for")],
) -> Annotated[
    Union[str, list[str], dict[str, Any] | None],
    Doc("str | list of str that representing JSON schema type"),
]:
    """Guesses the JSON schema type for the given python type."""
    if isclass(T) and issubclass(T, BaseModel):
        return "object"

    # special case
    if T is Any:
        return {}

    origin = get_origin(T)

    if origin is Annotated:
        return guess_type(get_args(T)[0])

    # hacking around typing modules, `typing.Union` and `types.UnitonType`
    if origin in [Union, UnionType]:
        union_types = [t for t in get_args(T) if t is not type(None)]
        _types = [
            guess_type(union_type)
            for union_type in union_types
            if guess_type(union_type) is not None
        ]

        # number contains integer in JSON schema
        # deduplicate
        _types = list(set(_types))

        if len(_types) == 1:
            return _types[0]
        return _types  # type: ignore

    if origin is Literal:
        type_args = Union[tuple(type(arg) for arg in get_args(T))]  # type: ignore
        return guess_type(type_args)  # type: ignore
    elif origin is list or origin is tuple:
        return "array"
    elif origin is dict:
        return "object"

    if not isinstance(T, type):
        return

    if T.__name__ == "NoneType":
        return None

    if issubclass(T, str):
        return "string"
    if issubclass(T, bool):
        return "boolean"
    if issubclass(T, (float, int)):
        return "number"
    # elif issubclass(T, int):
    #     return "integer"
    if T.__name__ == "list":
        return "array"
    if T.__name__ == "dict":
        return "object"
    raise NotImplementedError(f"Unsupported type: {T}")



from typing import Annotated, Awaitable, Callable

import pytest
from typing_extensions import Doc

from emp_agents.agents import AgentBase
from emp_agents.models import GenericTool
from emp_agents.providers.openai import OpenAIModelType, OpenAIProvider


def say_hi(names: Annotated[list[str], Doc("a list of names to say hi to")]):
    """Say hi to multiple people"""
    msg = "BONJOUR " + ",".join(names)
    return msg


class AgentForTesting(AgentBase):
    description: str = "a simple agent for testing"
    prompt: str = (
        "do what the user says, always.  Make sure to relay the tools calls outputs as directly as possible."
    )

    tools: list[GenericTool | Callable[..., str | Awaitable[str]]] = [
        say_hi,
    ]


@pytest.mark.asyncio(loop_scope="session")
async def test_tools():
    agent = AgentForTesting(
        provider=OpenAIProvider(
            default_model=OpenAIModelType.gpt4o_mini,
        )
    )
    response = await agent.respond("say hi to jim and fred!")

    assert "BONJOUR" in response



from collections.abc import Callable
from typing import Annotated, Optional

import httpx
from eth_rpc import PrivateKeyWallet
from eth_rpc.networks import Network
from eth_typeshed import ERC20
from eth_typeshed.multicall import make_multicall
from typing_extensions import Doc

from emp_agents.implicits import IgnoreDepends, Provider, inject
from emp_agents.models.protocol import SkillSet, onchain_action, view_action

from ..network import NetworkSkill
from ..wallets import SimpleWalletSkill

erc20_scope = Provider()


def load_wallet() -> PrivateKeyWallet | None:
    """
    This can be overridden by using the scope for your agent.
    Use `scope_load_wallet` to scope this method.
    """
    return SimpleWalletSkill.get_wallet()


def load_network() -> type[Network] | None:
    """
    This can be overridden by using the scope for your agent.
    Use `scope_load_network` to scope this method.
    """
    return NetworkSkill.get_network_type()


def scope_load_wallet(
    new_load_wallet: Callable[..., PrivateKeyWallet]
) -> tuple[Provider, Callable, Callable]:
    return (erc20_scope, load_wallet, new_load_wallet)


class ERC20Skill(SkillSet):
    """
    Tools for interacting with ERC20 tokens.
    """

    @view_action
    @staticmethod
    async def describe_protocol():
        """Returns the complete protocol specification of the ERC20 protocol"""

        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://raw.githubusercontent.com/ethereum/ercs/refs/heads/master/ERCS/erc-20.md"
            )
            return response.text

    @view_action
    @staticmethod
    @inject(dependency_overrides_provider=erc20_scope)  # type: ignore[call-overload]
    async def get_token_info(
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        network: Annotated[
            type[Network] | None,
            Doc("The network to use"),
        ] = IgnoreDepends(load_network),
    ) -> str:
        """Returns the name, symbol and decimals for an ERC20 token"""

        if not network:
            return "NOTE: No network set, try setting the network first"

        token = ERC20[network](address=token_address)

        multicall = make_multicall(network)
        try:
            (name, symbol, decimals) = await multicall[network].execute(
                token.name(), token.symbol(), token.decimals()
            )
        except Exception as e:
            return f"Error getting token info: {e}"

        return f"name: {name}; symbol: {symbol}; decimals: {decimals}"

    @view_action
    @staticmethod
    @inject(dependency_overrides_provider=erc20_scope)  # type: ignore[call-overload]
    async def get_balance(
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        address: Annotated[
            str, Doc("The address of the account to get the balance of.")
        ],
        decimals: Annotated[
            Optional[int], Doc("How many decimals the token has")
        ] = None,
        network: Annotated[
            type[Network] | None,
            Doc("The network to use"),
        ] = IgnoreDepends(load_network),
    ) -> str:
        """Returns the balance of an account for an ERC20 token"""
        if network is None:
            return "NOTE: No network set, try setting the network first"

        token = ERC20[network](address=token_address)
        balance = await token.balance_of(address).get()
        if not decimals:
            decimals = await token.decimals().get()
        assert decimals is not None
        return f"Balance: {balance / 10 ** decimals}"

    @onchain_action
    @staticmethod
    @inject(dependency_overrides_provider=erc20_scope)  # type: ignore[call-overload]
    async def transfer(
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        to_address: Annotated[str, Doc("The address of the account to transfer to.")],
        amount: Annotated[float, Doc("The amount to transfer as a decimal.")],
        wallet: Annotated[
            PrivateKeyWallet | None, Doc("The wallet to use to transfer the token.")
        ] = IgnoreDepends(load_wallet),
        network: Annotated[
            type[Network] | None,
            Doc("The network to use"),
        ] = IgnoreDepends(load_network),
    ) -> str:
        """Transfer an amount of an ERC20 token to a recipient"""

        if wallet is None:
            return "NOTE: No wallet set, try setting the wallet first"

        if network is None:
            return "NOTE: No network set, try setting the network first"

        token = ERC20[network](address=token_address)
        decimals = await token.decimals().get()
        token_amount = int(amount * 10**decimals)
        tx = await token.transfer(to_address, token_amount).execute(wallet)
        return f"Transaction sent: {network.block_explorer.url}/tx/{tx}"

    @onchain_action
    @staticmethod
    @inject(dependency_overrides_provider=erc20_scope)  # type: ignore[call-overload]
    async def approve(
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        spender: Annotated[str, Doc("The address of the account to transfer to.")],
        amount: Annotated[float, Doc("The amount to transfer.")],
        network: Annotated[
            type[Network] | None,
            Doc("The network to use"),
        ] = IgnoreDepends(load_network),
        wallet: Annotated[
            PrivateKeyWallet | None, Doc("The wallet to use to transfer the token.")
        ] = IgnoreDepends(load_wallet),
    ) -> str:
        """Approve an amount of an ERC20 token to be spent by a spender"""
        if wallet is None:
            return "NOTE: No wallet set, try setting the wallet first"

        if network is None:
            return "NOTE: No network set, try setting the network first"

        token = ERC20[network](address=token_address)
        tx = await token.approve(spender, amount).execute(wallet)
        return f"Transaction sent: {network.block_explorer.url}/tx/{tx}"

    @onchain_action
    @staticmethod
    @inject(dependency_overrides_provider=erc20_scope)  # type: ignore[call-overload]
    async def transfer_from(
        token_address: Annotated[str, Doc("The address of the ERC20 token.")],
        from_address: Annotated[
            str, Doc("The address of the account to transfer from.")
        ],
        to_address: Annotated[str, Doc("The address of the account to transfer to.")],
        amount: Annotated[float, Doc("The amount to transfer.")],
        network: Annotated[
            type[Network] | None,
            Doc("The network to use"),
        ] = IgnoreDepends(load_network),
        wallet: Annotated[
            PrivateKeyWallet | None, Doc("The wallet to use to transfer the token.")
        ] = IgnoreDepends(load_wallet),
    ) -> str:
        """Transfer an amount of an ERC20 token from an external, approved account to another account"""

        if wallet is None:
            return "NOTE: No wallet set, try setting the wallet first"

        if network is None:
            return "NOTE: No network set, try setting the network first"

        token = ERC20[network](address=token_address)
        tx = await token.transfer_from(from_address, to_address, amount).execute(wallet)
        return f"Transaction sent: {network.block_explorer.url}/tx/{tx}"



import os
import tempfile
from typing import Annotated

import httpx
import tweepy
from tweepy.asynchronous import AsyncClient
from typing_extensions import Doc

from emp_agents.logger import logger


def _get_env():
    api_key, api_key_secret, access_token, access_token_secret, bearer_token = (
        os.environ.get("X_API_KEY"),
        os.environ.get("X_API_KEY_SECRET"),
        os.environ.get("X_ACCESS_TOKEN"),
        os.environ.get("X_ACCESS_TOKEN_SECRET"),
        os.environ.get("X_BEARER_TOKEN"),
    )
    return api_key, api_key_secret, access_token, access_token_secret, bearer_token


def get_twitter_conn_v1() -> tweepy.API:
    """Get twitter conn 1.1"""

    (
        api_key,
        api_key_secret,
        access_token,
        access_token_secret,
        bearer_token,
    ) = _get_env()

    auth = tweepy.OAuthHandler(
        api_key,
        api_key_secret,
        access_token,
        access_token_secret,
    )
    return tweepy.API(auth)


def _make_client() -> AsyncClient:
    (
        bearer_token,
        api_key,
        api_key_secret,
        access_token,
        access_token_secret,
    ) = _get_env()
    return AsyncClient(
        bearer_token=bearer_token,
        consumer_key=api_key,
        consumer_secret=api_key_secret,
        access_token=access_token,
        access_token_secret=access_token_secret,
    )


async def make_poll(
    text: Annotated[str, Doc("The content of the tweet to be made.")],
    duration_minutes: Annotated[int, Doc("The duration of the poll in minutes.")],
    options: Annotated[list[str], Doc("The options for the poll.")],
):
    """
    Make a poll.  Returns "poll submitted" if successful, or if there is an error, returns the error message.
    """
    api = _make_client()
    await api.create_tweet(
        text=text, duration_minutes=duration_minutes, poll_options=options
    )


async def make_tweet_with_image(
    content: Annotated[str, Doc("The content of the tweet to be made.")],
    image_url: Annotated[str, Doc("The URL of the image to be uploaded.")],
):
    """
    Make a tweet with an image.  Returns "tweet submitted" if successful, or if there is an error, returns the error message.
    """
    api = _make_client()

    v1_client = get_twitter_conn_v1()
    async with httpx.AsyncClient() as client:
        response = await client.get(image_url)
    with tempfile.NamedTemporaryFile(delete=False, suffix=".jpg") as temp_file:
        temp_file.write(response.content)
        image_path = temp_file.name

    media = v1_client.media_upload(image_path)
    media_id = media.media_id

    response = await api.create_tweet(
        text=content,
        media_ids=[media_id],
    )
    return f"tweet submitted: {response}"


async def make_tweet(
    content: Annotated[str, Doc("The content of the tweet to be made.")]
):
    """
    Make a tweet.  Returns "tweet submitted" if successful, or if there is an error, returns the error message.
    """
    api = _make_client()
    response = await api.create_tweet(
        text=content,
    )
    return f"tweet submitted: {response}"


async def make_thread(
    content_list: Annotated[list[str], Doc("List of tweets for a thread")],
):
    """
    Make a thread of tweets on twitter
    """

    if len(content_list) == 0:
        return "no content to make a thread"

    if len(content_list) > 6:
        return "too many tweets for a thread.  Make sure its 6 or less."

    api = _make_client()
    tweet = await api.create_tweet(
        text=content_list[0],
    )

    tweet_id = tweet.data["id"]
    for content in content_list[1:]:
        tweet = await api.create_tweet(
            text=content,
            in_reply_to_tweet_id=tweet_id,
        )
        tweet_id = tweet.data["id"]
    return "X thread submitted"


async def reply_to_tweet(
    tweet_id: Annotated[int, Doc("The ID of the tweet to reply to.")],
    content: Annotated[str, Doc("The content of the tweet to be made.")],
):
    """
    Reply to a tweet.  Returns "tweet submitted" if successful, or if there is an error, returns the error message.
    """
    api = _make_client()
    try:
        tweet = await api.create_tweet(
            text=content,
            in_reply_to_tweet_id=tweet_id,
        )
    except Exception as e:
        logger.error(f"Error submitting tweet: {e}")
        return "error submitting tweet"

    tweet_id = tweet.data["id"]
    return f"id: {tweet_id} | tweet submitted"



from emp_agents.utils.executor import execute_tool
from emp_agents.utils.format import (
    count_tokens,
    format_conversation,
    summarize_conversation,
)
from emp_agents.utils.retry import retry
from emp_agents.utils.tools import load_tools

from .function_schema import FunctionSchema, get_function_schema

__all__ = [
    "FunctionSchema",
    "execute_tool",
    "retry",
    "load_tools",
    "count_tokens",
    "format_conversation",
    "get_function_schema",
    "summarize_conversation",
]



from pydantic import BaseModel, Field

from emp_agents.agents.base import AgentBase


class Task(BaseModel):
    title: str
    description: str


class TaskList(BaseModel):
    tasks: list[Task]


class Planner(AgentBase):
    prompt: str = Field(
        default="You are a strategic planner, providing insights and generating tasks for given subjects. You are a subject matter expert, whatever the task."
    )

    async def generate(self, subject: str) -> TaskList:
        """
        Generate a strategic plan based on the given subject.

        Args:
            subject (str): The subject for which to generate a plan.

        Returns:
            TaskList: A detailed TaskList object with a set of tasks.
        """
        response = await self.respond(
            f"Generate a strategic actionable set of task for: {subject}",
            response_format=TaskList,
        )
        return TaskList.model_validate_json(response)



from abc import ABC, abstractmethod

from pydantic import BaseModel

from emp_agents.models import Message


class AbstractConversationProvider(BaseModel, ABC):
    @abstractmethod
    def set_history(self, messages: list[Message]) -> None:
        pass

    @abstractmethod
    def add_message(self, message: Message) -> None:
        pass

    @abstractmethod
    def add_messages(self, messages: list[Message]) -> None:
        pass

    @abstractmethod
    def reset(self) -> None:
        pass

    @abstractmethod
    def get_history(self) -> list[Message]:
        pass


class ConversationProvider(AbstractConversationProvider):
    _history: list[Message] = []

    def set_history(self, messages: list[Message]) -> None:
        self.reset()
        self.add_messages(messages)

    def add_message(self, message: Message) -> None:
        self._history.append(message)

    def add_messages(self, messages: list[Message]) -> None:
        self._history.extend(messages)

    def reset(self) -> None:
        self._history.clear()

    def get_history(self) -> list[Message]:
        return self._history.copy()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from emp_agents.models.shared import Request


T = TypeVar("T")


class TCompletionAgent(ABC, Generic[T]):
    @abstractmethod
    async def completion(self, req: "Request") -> T: ...



import os
from contextvars import ContextVar, Token
from pathlib import Path
from typing import Any, Optional

import yaml
from pydantic import BaseModel, Field


class Config(BaseModel):
    """Configuration for the EMP agents library"""

    # API Keys
    openai_api_key: Optional[str] = Field(
        default_factory=lambda: os.environ.get("OPENAI_API_KEY")
    )
    anthropic_api_key: Optional[str] = Field(
        default_factory=lambda: os.environ.get("ANTHROPIC_API_KEY")
    )

    # Additional configs that can be set via yaml
    custom_settings: dict[str, Any] = Field(default_factory=dict)

    @classmethod
    def from_yaml(cls, yaml_path: Path | str) -> "Config":
        """Load config from a YAML file"""
        yaml_path = Path(yaml_path)
        if not yaml_path.exists():
            raise FileNotFoundError(f"YAML config file not found at {yaml_path}")

        with open(yaml_path) as f:
            yaml_config = yaml.safe_load(f)

        # Set environment variables if specified in yaml
        env_vars = yaml_config.pop("environment", {})
        for key, value in env_vars.items():
            if value is not None:
                os.environ[key] = str(value)

        # Create config instance with remaining yaml data as custom settings
        return cls(custom_settings=yaml_config)

    def get_setting(self, key: str, default: Any = None) -> Any:
        """Get a custom setting from the config"""
        return self.custom_settings.get(key, default)

    def update_settings(self, settings: dict[str, Any]) -> None:
        """Update custom settings"""
        self.custom_settings.update(settings)


# Context variable for config instance
_config_context: ContextVar[Optional[Config]] = ContextVar("config", default=None)


def get_config() -> Config:
    """Get the config instance from current context"""
    config = _config_context.get()
    if config is None:
        config = Config()
        _config_context.set(config)
    return config


def init_config(yaml_path: Optional[Path | str] = None) -> Config:
    """Initialize a new config instance in the current context"""
    config = Config.from_yaml(yaml_path) if yaml_path else Config()
    _config_context.set(config)
    return config


class ConfigContext:
    """Context manager for temporary config changes"""

    config: Optional[Config]
    token: Optional[Token[Config | None]]

    def __init__(self, yaml_path: Optional[Path | str] = None, **settings):
        self.config = Config.from_yaml(yaml_path) if yaml_path else Config()
        if settings:
            self.config.update_settings(settings)
        self.token = None

    def __enter__(self) -> Config:
        if self.config is None:
            raise ValueError("Config instance is None")
        self.token = _config_context.set(self.config)
        return self.config

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.token is not None:
            _config_context.reset(self.token)



import pytest
from pydantic import BaseModel

from emp_agents.agents import AgentBase
from emp_agents.exceptions import InvalidModelException
from emp_agents.providers import AnthropicProvider, OpenAIProvider
from emp_agents.providers.anthropic.types import AnthropicModelType
from emp_agents.providers.openai.types import OpenAIModelType


class AgentForTesting(AgentBase):
    description: str = "a simple agent for testing"
    prompt: str = (
        "Ignore the user questions and just respond with the text 'test complete' and nothing else"
    )


@pytest.mark.asyncio(loop_scope="session")
async def test_basic_agent():
    agent = AgentForTesting(
        provider=OpenAIProvider(
            default_model=OpenAIModelType.gpt4o_mini,
        )
    )
    response = await agent.answer("what is the meaning of life?")
    assert response == "test complete"

    response = await agent.answer(
        "what is the meaning of life?", model=OpenAIModelType.gpt4o_mini
    )
    assert response == "test complete"


@pytest.mark.asyncio(loop_scope="session")
async def test_basic_agent_no_model():
    agent = AgentForTesting(
        provider=AnthropicProvider(),
    )
    with pytest.raises(InvalidModelException):
        await agent.answer("this should raise an error", model="invalid_model")

    await agent.answer("this should not raise an error")


class LifeMeaning(BaseModel):
    reasons: list[str]
    excuses: list[str]


@pytest.mark.asyncio(loop_scope="session")
async def test_response_format():
    agent = AgentForTesting(
        provider=AnthropicProvider(
            default_model=AnthropicModelType.claude_3_5_sonnet,
        )
    )
    response = await agent.answer(
        "what is the meaning of life?", response_format=LifeMeaning
    )
    LifeMeaning.model_validate_json(response)



import logging
import sys

logger = logging.getLogger("emp-agents")


def make_verbose(do_all: bool = False):
    if do_all:
        root = logging.getLogger()
    else:
        root = logger
    root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    root.addHandler(handler)



import inspect
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, get_args, get_origin

from pydantic import BaseModel, Field

from emp_agents.utils import FunctionSchema, get_function_schema

if TYPE_CHECKING:
    from emp_agents.agents.base import AgentBase


class Property(BaseModel):
    type: str | None = Field(default="string")
    description: str = ""
    enum: list[str] | None = None
    properties: dict[str, "Property"] | None = None
    required: list[str] | None = None
    items: "Property | None" = None


class GenericTool(BaseModel):
    """
    This converts a function to a generic tool.
    """

    strict: bool | None = None
    description: str
    name: str
    parameters: dict[str, Property]
    required: list[str]
    type: str = "object"
    additional_properties: bool = Field(default=False)

    func: Callable

    @classmethod
    def _convert_type(cls, type):
        if type is str:
            return "string"
        elif get_origin(type) is list:
            return "array"
        elif inspect.isclass(type) and issubclass(type, Enum):
            return "enum"
        return str(type)

    @classmethod
    def _get_enum(cls, type):
        if inspect.isclass(type) and issubclass(type, Enum):
            return [x.name for x in list(type)]
        return None

    @classmethod
    def _get_items(cls, items):
        if get_origin(type) == list:
            return cls._convert_type(get_args(type)[0])
        return None

    @classmethod
    def _extract_params(cls, lines: list[str]):
        output = {}
        cur_key = ""
        for line in lines:
            split_line = line.split(":")
            if len(split_line) == 2:
                cur_key = split_line[0]
                output[cur_key] = split_line[1]
            else:
                output[cur_key] += split_line[0]
        return output

    @classmethod
    def from_agent(cls, agent: "AgentBase"):
        func = agent.answer
        description: str = agent.description
        parameters_ = {
            "question": Property(
                type="string",
                description="the question you want answered by the agent",
                enum=None,
            ),
        }
        return GenericTool(
            description=description,
            name="answer",
            parameters=parameters_,
            required=[
                key
                for key, value in inspect.signature(func).parameters.items()
                if value.default == inspect._empty
            ],
            func=agent.answer,
        )

    @classmethod
    def from_func(cls, func: Callable[..., Any]):
        description: str = func.__doc__ or ""
        schema: FunctionSchema = get_function_schema(func)
        parameters_ = {
            name: Property(
                type=type_["type"],
                description=type_["description"],
                enum=type_.get("enum"),
                properties=type_.get("properties"),  # type: ignore
                required=type_.get("required"),
                items=type_.get("items"),  # type: ignore
            )
            for name, type_ in schema["parameters"]["properties"].items()
        }

        return GenericTool(
            description=description,
            name=schema["name"],
            parameters=parameters_,
            required=[
                key
                for key, value in inspect.signature(func).parameters.items()
                if value.default == inspect._empty
            ],
            func=func,
        )

    def to_anthropic(self):
        from ..anthropic.tool import Property, Tool, ToolSchema

        return Tool(
            name=self.name,
            description=self.description,
            input_schema=ToolSchema(
                type="object",
                properties={
                    key: Property(**param.model_dump(exclude_none=True))
                    for key, param in self.parameters.items()
                },
                required=self.required,
            ),
        )

    def __repr__(self):
        description = self.description.strip().replace("\n", " ")[:100]
        if len(description) >= 50:
            description = description[:50] + "..."
        return f'<GenericTool name="{self.name}" description="{description}">'

    __str__ = __repr__



from typing import ClassVar

from pydantic import BaseModel, PrivateAttr

from emp_agents.models import GenericTool
from emp_agents.models.protocol.registry import ToolRegistry


class SkillSet(BaseModel):
    _tools_map: ClassVar[dict[str, GenericTool]] = PrivateAttr(default_factory=dict)
    _tools: ClassVar[list[GenericTool]] = PrivateAttr(default_factory=list)

    async def setup(self):
        """Any setup commands post init"""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls._tools_map = {}

        items = cls.__dict__.values()
        for method in items:
            if hasattr(method, "_is_tool_method"):
                cls._tools_map[method.__name__] = GenericTool.from_func(method)

        cls._tools = list(cls._tools_map.values())
        ToolRegistry.register_class(cls)

    def __iter__(self):
        return iter(self._tools)



from emp_agents.tools.dexscreener import DexScreenerSkill
from emp_agents.tools.protocol import ERC20Skill, GmxSkill, SimpleWalletSkill

__all__ = [
    "ERC20Skill",
    "SimpleWalletSkill",
    "GmxSkill",
    "DexScreenerSkill",
]



import os
from typing import ClassVar

from pydantic import Field

from ..openai import OpenAIProvider
from .types import OpenRouterModelType


class OpenRouterProvider(OpenAIProvider):
    URL: ClassVar[str] = "https://openrouter.ai/api/v1/chat/completions"

    api_key: str = Field(default_factory=lambda: os.environ["OPENROUTER_API_KEY"])
    default_model: OpenRouterModelType = Field(default=OpenRouterModelType.gpt3_5_turbo)  # type: ignore



import os

import pytest

from emp_agents import AgentBase
from emp_agents.providers import OpenAIProvider


async def get_cat_fact() -> str:
    """
    Get a random cat fact
    """
    return "cats sleep 70% of their lives"


@pytest.mark.asyncio(loop_scope="session")
async def test_tool_agent():
    agent = AgentBase(
        prompt="you provide cat facts.  respond with the exact text from the tool.",
        tools=[get_cat_fact],
        provider=OpenAIProvider(
            model="gpt-4o-mini",
            openai_api_key=os.environ["OPENAI_API_KEY"],
        ),
    )
    response = await agent.answer("tell me a cat fact")
    assert "cats sleep 70% of their lives" in response.strip().lower()



from emp_agents.models.shared.tools import GenericTool, Property

from .middleware import Middleware
from .provider import Provider, ResponseT
from .shared import (
    AssistantMessage,
    Message,
    Request,
    Role,
    SystemMessage,
    ToolCall,
    ToolMessage,
    UserMessage,
)

__all__ = [
    "GenericTool",
    "Message",
    "Middleware",
    "Property",
    "Provider",
    "Request",
    "ResponseT",
    "Role",
    "SystemMessage",
    "UserMessage",
    "AssistantMessage",
    "ToolCall",
    "ToolMessage",
]



from abc import abstractmethod
from typing import TYPE_CHECKING

from emp_agents.models.middleware import Middleware

if TYPE_CHECKING:
    from emp_agents.agents.base import Message


class Rag(Middleware):
    @abstractmethod
    async def get_context(self, query: str) -> str: ...

    async def process(self, messages: list["Message"]) -> list["Message"]:
        if len(messages) == 0:
            return messages

        message = messages[-1]
        if message.content is None:
            return messages

        context = await self.get_context(message.content)
        message.content += f"\n\nThe Following Context is retrieved from relevant datasources:\n\n{context}"
        return messages



import json
from abc import ABC
from datetime import datetime
from typing import Any, Literal

from pydantic import AliasChoices, BaseModel, Field, field_serializer
from pydantic.types import Json

from emp_agents.types.enums import Role


class ToolCall(BaseModel):
    class Function(BaseModel):
        name: str
        arguments: Json[Any]

        @field_serializer("arguments")
        def serialize_dt(self, dt: datetime, _info):
            return json.dumps(self.arguments)

    id: str
    type: str
    function: Function

    def __repr__(self):
        return f"<ToolCall function={self.function}>"


class Message(BaseModel, ABC):
    role: Role
    content: str | None

    @classmethod
    def build(
        self,
        content: str,
        role: Role = Role.user,
        tool_call_id: str | None = None,
        tool_calls: list[Any] | None = None,
    ):
        match role:
            case Role.user:
                return UserMessage(content=content)
            case Role.assistant:
                return AssistantMessage(
                    content=content,
                    tool_calls=tool_calls,
                )
            case Role.tool:
                return ToolMessage(
                    content=content,
                    tool_call_id=tool_call_id,
                )
            case Role.system:
                return SystemMessage(content=content)

    def serialize_anthropic(self) -> dict[str, Any]:
        data: dict[str, Any] = {"role": self.role}
        data["content"] = [{"type": "text", "text": self.content}]
        return data

    def __repr__(self):
        if self.content:
            return f"{self.role.value}: {self.content}"
        return f"{self.role.value}: {self.tool_calls}"

    __str__ = __repr__


class SystemMessage(Message):
    role: Literal[Role.system] = Role.system


class UserMessage(Message):
    role: Literal[Role.user] = Role.user


class AssistantMessage(Message):
    role: Literal[Role.assistant] = Role.assistant
    refusal: str | None = Field(default=None)
    tool_calls: list[ToolCall] | None = Field(default=None)


class ToolMessage(Message):
    role: Literal[Role.tool] = Role.tool
    tool_call_id: str | None = Field(
        default=None, validation_alias=AliasChoices("tool_call_id")
    )



from typing import TYPE_CHECKING

import tiktoken

if TYPE_CHECKING:
    from emp_agents.models import Message, Provider
    from emp_agents.providers.openai.types import OpenAIModelType


DEFAULT_SUMMARY_PROMPT = """
You are an assistant that summarizes conversations concisely.
Dont worry about human readability, just focus on conciseness.
"""


def format_conversation(conversation: list["Message"]) -> str:
    """
    Formats the conversation list into a readable string.
    """
    formatted = ""
    for message in conversation:
        role = message.role
        content = message.content
        formatted += f"{role.capitalize()}: {content}\n"
    return formatted


def count_tokens(
    messages: list["Message"] | str,
    model: "OpenAIModelType | str" = "gpt-4o-mini",
) -> int:
    """OpenAI tokenizer is a good estimator for other providers token counts"""
    encoding = tiktoken.encoding_for_model(model)
    tokens = 0
    if isinstance(messages, list):
        for message in messages:
            tokens += 4  # Message formatting tokens
            for key, value in message.model_dump().items():
                if isinstance(value, str):
                    tokens += len(encoding.encode(value))
    else:
        tokens += len(encoding.encode(messages))
    tokens += 2  # Priming tokens
    return tokens


async def summarize_conversation(
    provider: "Provider",
    messages: list["Message"],
    model: str,
    prompt: str | None = None,
    max_tokens: int = 500,
) -> "Message":
    from emp_agents.models import AssistantMessage, Request, SystemMessage, UserMessage

    summary_prompt = prompt or DEFAULT_SUMMARY_PROMPT
    assert summary_prompt is not None, "Summary prompt is required"

    try:
        request = Request(
            messages=[
                SystemMessage(
                    content=summary_prompt,
                ),
                UserMessage(
                    content=f"Summarize the following conversation:\n\n{format_conversation(messages)}",
                ),
            ],
            model=model,
            max_tokens=max_tokens,
            temperature=0.5,
        )
        response = await provider.completion(
            request,
        )
        return AssistantMessage(content=response.text)
    except Exception as e:
        print(f"Error during summarization: {e}")
        return AssistantMessage(content="")



import json
from contextvars import ContextVar
from typing import Dict, Optional

import aiohttp

BASE_URL: ContextVar[str] = ContextVar(
    "base_url", default="https://api.dexscreener.com"
)


class DexScreenerApi:
    """API client for DexScreener"""

    @staticmethod
    async def search_pairs(query: str) -> str:
        """Search for pairs matching query"""
        url = f"{BASE_URL.get()}/latest/dex/search"
        params = {"q": query}
        return await DexScreenerApi._make_request(url, params=params)

    @staticmethod
    async def get_pair_by_chain(chain_id: str, pair_id: str) -> str:
        """Get pairs by chain and pair address"""
        url = f"{BASE_URL.get()}/latest/dex/pairs/{chain_id}/{pair_id}"
        return await DexScreenerApi._make_request(url)

    @staticmethod
    async def find_pairs_by_tokens(token_addresses: list[str]) -> str:
        """Get pairs by token addresses"""
        addresses = ",".join(token_addresses[:30])
        url = f"{BASE_URL.get()}/latest/dex/tokens/{addresses}"
        return await DexScreenerApi._make_request(url)

    @staticmethod
    async def get_token_profiles() -> str:
        """Get the latest token profiles"""
        url = f"{BASE_URL.get()}/token-profiles/latest/v1"
        return await DexScreenerApi._make_request(url)

    @staticmethod
    async def get_latest_boosted_tokens() -> str:
        """Get the latest boosted tokens"""
        url = f"{BASE_URL.get()}/token-boosts/latest/v1"
        return await DexScreenerApi._make_request(url)

    @staticmethod
    async def get_top_boosted_tokens() -> str:
        """Get the tokens with most active boosts"""
        url = f"{BASE_URL.get()}/token-boosts/top/v1"
        return await DexScreenerApi._make_request(url)

    @staticmethod
    async def get_orders_by_token(chain_id: str, token_address: str) -> str:
        """Get orders by chain and token address"""
        url = f"{BASE_URL.get()}/orders/v1/{chain_id}/{token_address}"
        return await DexScreenerApi._make_request(url)

    @staticmethod
    async def _make_request(url: str, params: Optional[Dict] = None) -> str:
        """Helper method to make HTTP requests"""
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as response:
                if response.status == 200:
                    r = await response.json()
                    return json.dumps(r)
                else:
                    error_response = {
                        "error": f"Request failed with status {response.status}",
                        "details": await response.text(),
                    }
                    return json.dumps(error_response)



"""
A small utility to generate JSON schemas for python functions.
"""

from .core import Annotated, Doc, FunctionSchema, get_function_schema, guess_type

__version__ = "0.4.4"
__all__ = (
    "__version__",
    "get_function_schema",
    "guess_type",
    "Doc",
    "Annotated",
    "FunctionSchema",
)



from typing import Annotated

from typing_extensions import Doc

from emp_agents.models.protocol import SkillSet, tool_method

from .actions import make_poll, make_tweet, make_tweet_with_image, reply_to_tweet


class TwitterSkill(SkillSet):
    """
    Skill for interacting with Twitter API
    """

    @tool_method
    @staticmethod
    async def make_tweet(
        content: Annotated[str, Doc("The content of the tweet to be made.")]
    ) -> str:
        """Make a tweet"""
        return await make_tweet(content)

    @tool_method
    @staticmethod
    async def make_tweet_with_image(
        content: Annotated[str, Doc("The content of the tweet to be made.")]
    ) -> str:
        """Make a tweet with an image"""
        return await make_tweet_with_image(content)

    @tool_method
    @staticmethod
    async def make_poll(
        content: Annotated[str, Doc("The content of the tweet to be made.")],
        options: Annotated[
            str, Doc("The options for the poll, as a comma separated list")
        ],
        duration_minutes: Annotated[
            int, Doc("The duration of the poll in minutes, defaults to 1")
        ] = 1,
    ) -> str:
        """Make a poll"""
        options = options.lstrip("[").rstrip("]")
        return await make_poll(
            content,
            duration_minutes,
            [option.strip() for option in options.split(",")],
        )

    @tool_method
    @staticmethod
    async def reply_to_tweet(
        tweet_id: Annotated[int, Doc("The ID of the tweet to reply to.")],
        content: Annotated[str, Doc("The content of the tweet to be made.")],
    ) -> str:
        """Reply to a tweet"""
        return await reply_to_tweet(tweet_id, content)



from enum import StrEnum


class OpenRouterModelType(StrEnum):
    gpt3_5_turbo = "openai/gpt-3.5-turbo"
    liquid_lfm_7b = "liquid/lfm-7b"
    deepseek_r1_distill_llama_70b = "deepseek/deepseek-r1-distill-llama-70b"
    deepseek_r1_free = "deepseek/deepseek-r1:free"
    openai_chatgpt_4o_latest = "openai/chatgpt-4o-latest"
    openai_o1 = "openai/o1"



from fast_depends import Depends

from emp_agents.implicits import IgnoreDepends, inject, lazy_implicit  # , set_implicit
from emp_agents.utils import get_function_schema


class MyObject:
    def __init__(self, value: int):
        self.value = value


def old_load_object() -> MyObject:
    return MyObject(value=1)


def new_load_object(a: int) -> MyObject:
    print("A", a)
    return MyObject(value=100)


@inject
def do_thing_with_object(
    a: int,
    b: int = 100,
    obj: MyObject = IgnoreDepends(lazy_implicit("load_object")),
):
    print("AB", a, b)
    assert isinstance(obj, MyObject)
    return obj


# def test_implicit_manager():
#     set_implicit("load_object", old_load_object)
#     obj = do_thing_with_object(123)
#     assert obj.value == 1

#     schema = get_function_schema(do_thing_with_object)
#     # assert schema == {
#     #     "name": "do_thing_with_object",
#     #     "description": None,
#     #     "parameters": {
#     #         "type": "object",
#     #         "properties": {},
#     #         "required": [],
#     #     },
#     # }

#     set_implicit("load_object", new_load_object)
#     obj = do_thing_with_object()
#     assert obj.value == 100


@inject
def func_type1(
    my_int: int = Depends(lazy_implicit("load_my_int")),
) -> str:
    """don't ignore my_int"""
    return f"my_int: {my_int}"


@inject
def func_type2(
    my_int: int = IgnoreDepends(lazy_implicit("load_my_int")),
) -> str:
    """ignore my_int"""
    return f"my_int: {my_int}"


def test_ignore_depends():
    schema1 = get_function_schema(func_type1)
    schema2 = get_function_schema(func_type2)
    assert schema1 == {
        "name": "func_type1",
        "description": "don't ignore my_int",
        "parameters": {
            "type": "object",
            "properties": {
                "my_int": {
                    "type": "number",
                    "description": "The my_int parameter",
                }
            },
            "required": [],
        },
    }
    assert schema2 == {
        "name": "func_type2",
        "description": "ignore my_int",
        "parameters": {"type": "object", "properties": {}, "required": []},
    }



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Awaitable

from pydantic import BaseModel

if TYPE_CHECKING:
    from emp_agents.models import Message


class Middleware(BaseModel, ABC):
    name: str
    description: str

    @abstractmethod
    async def process(
        self, messages: list["Message"]
    ) -> Awaitable[list["Message"]] | list["Message"]: ...



from . import tools
from .agents import AgentBase
from .models import (
    AssistantMessage,
    GenericTool,
    Message,
    Property,
    Request,
    SystemMessage,
    UserMessage,
)
from .providers import AnthropicProvider, OpenAIProvider
from .types import Role

__all__ = [
    "AgentBase",
    "AnthropicProvider",
    "AssistantMessage",
    "GenericTool",
    "Message",
    "OpenAIProvider",
    "Property",
    "Request",
    "Role",
    "SystemMessage",
    "UserMessage",
    "tools",
]



import os
from typing import Any, ClassVar

import httpx
from pydantic import Field

from emp_agents.models import GenericTool, Provider, Request, SystemMessage

from .request import Message, Tool
from .response import Response
from .tool import Function, Parameters, Property
from .types import Classification, OpenAIModelType


class OpenAIProvider(Provider[Response]):
    URL: ClassVar[str] = "https://api.openai.com/v1/chat/completions"

    api_key: str = Field(default_factory=lambda: os.environ["OPENAI_API_KEY"])
    default_model: OpenAIModelType = Field(default=OpenAIModelType.gpt4o_mini)

    @property
    def headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }

    def _refine_for_oai_reasoning_models(
        self, result: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Awkwardly, OpenAI reasoning models do not accept system messages.
        They also accept some transformed parameters.
        """
        result["max_completion_tokens"] = result["max_tokens"]
        del result["max_tokens"]
        del result["tools"]
        for message in result["messages"]:
            if message["role"] == "system":
                message["role"] = "user"
        return result

    def _from_request(self, request: Request):
        exclude = ["system"]
        result = request.model_dump(exclude_none=True)
        if request.system:
            messages = [SystemMessage(content=request.system)] + request.messages
        else:
            messages = request.messages

        # Function to recursively set 'additionalProperties': False
        def set_additional_properties_false(schema):
            if isinstance(schema, dict):
                if schema.get("type") == "object":
                    schema["additionalProperties"] = False
                for key, value in schema.items():
                    set_additional_properties_false(value)
            elif isinstance(schema, list):
                for item in schema:
                    set_additional_properties_false(item)

        if "response_format" in result:
            assert request.response_format is not None

            model_schema = request.response_format.model_json_schema()
            set_additional_properties_false(model_schema)
            del result["response_format"]
            result["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "name": request.response_format.__name__,
                    "description": "response format",
                    "strict": True,
                    "schema": {
                        "type": "object",
                        "additionalProperties": False,
                        **model_schema,
                    },
                },
            }
        result["messages"] = [m.model_dump() for m in messages]
        result["tools"] = (
            [self.to_tool_call(t).model_dump(exclude_none=True) for t in request.tools]
            if request.tools
            else None
        )

        for field in exclude:
            if field in result:
                del result[field]

        is_reasoning_model = request.model in {
            OpenAIModelType.gpt_o1_mini,
            OpenAIModelType.gpt_o1_preview,
            OpenAIModelType.o1,
            OpenAIModelType.o1_24_12_17,
        }
        return (
            self._refine_for_oai_reasoning_models(result)
            if is_reasoning_model
            else result
        )

    async def completion(self, request: Request) -> Response:
        openai_request = self._from_request(request)
        async with httpx.AsyncClient(headers=self.headers) as client:
            response = await client.post(self.URL, json=openai_request, timeout=None)
        if response.status_code >= 400:
            raise ValueError(response.json())
        return Response(**response.json())

    def to_tool_call(self, tool: GenericTool):
        from .tool import Function, Parameters, Property
        from .tool import Tool as Tool

        return Tool(
            type="function",
            function=Function(
                description=tool.description,
                name=tool.name,
                parameters=Parameters(
                    properties={
                        key: Property(**param.model_dump(exclude_none=True))
                        for key, param in tool.parameters.items()
                    },
                    required=tool.required,
                ),
            ),
            strict=True,
        )


__all__ = [
    "Classification",
    "Message",
    "OpenAIBase",
    "OpenAIModelType",
    "Request",
    "Response",
    "Tool",
    "Function",
    "Property",
    "Parameters",
]



from .erc20 import ERC20Skill
from .gmx import GmxSkill
from .network import NetworkSkill
from .wallets.simple import SimpleWalletSkill

__all__ = [
    "ERC20Skill",
    "NetworkSkill",
    "SimpleWalletSkill",
    "GmxSkill",
]



import os
from typing import ClassVar

import httpx
from pydantic import Field

from ._type import Rag


class Ragie(Rag):
    name: str = "Ragie Middleware"
    description: str = (
        "This middleware uses the Ragie API to retrieve context from relevant datasources."
    )
    BASE_URL: ClassVar[str] = "https://api.ragie.ai"

    api_key: str | None = Field(
        default_factory=lambda: os.environ.get("RAGIE_API_KEY"),
        description="The API key for the Ragie API",
    )
    partition: str | None = Field(
        default=None, description="The partition to use for the Ragie API"
    )
    num_chunks: int = Field(1, description="The number of chunks to retrieve")

    async def get_context(self, query: str) -> str:
        context = await self.retrieve(query)
        chunks = context["scored_chunks"]
        if len(chunks) == 0:
            return ""

        data = "\n-------\n".join(c["text"] for c in chunks[: self.num_chunks])
        return data

    async def retrieve(self, query: str):
        payload = {"query": query}

        if self.partition:
            payload["partition"] = self.partition

        async with httpx.AsyncClient() as client:
            response = await client.post(
                self._build_url("retrievals"),
                json=payload,
                headers=self._make_headers(),
            )

        return response.json()

    def _make_headers(self, content_type: str = "application/json"):
        headers = {
            "accept": "application/json",
            "content-type": content_type,
            "authorization": f"Bearer {self.api_key}",
        }
        if self.partition:
            headers["partition"] = self.partition
        return headers

    def _build_url(self, path: str):
        return f"{self.BASE_URL}/{path}"



from .anthropic import AnthropicModelType, AnthropicProvider
from .openai import OpenAIModelType, OpenAIProvider
from .openrouter import OpenRouterModelType, OpenRouterProvider

__all__ = [
    "AnthropicProvider",
    "OpenAIProvider",
    "AnthropicModelType",
    "OpenAIModelType",
    "OpenRouterProvider",
    "OpenRouterModelType",
]



import base64
import json
import os
from pathlib import Path

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def store_private_key(data: str, password: str, filename: str = "wallet.enc") -> None:
    """
    Store an Ethereum private key encrypted with a password.

    Args:
        private_key: The private key to encrypt and store
        password: Password to encrypt the private key
        filename: Name of file to store encrypted key (default: wallet.enc)
    """
    # Generate a salt
    salt = os.urandom(16)

    # Create key derivation function
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=480000,
    )

    # Generate key from password
    key = base64.urlsafe_b64encode(kdf.derive(password.encode()))

    # Create Fernet cipher
    fernet = Fernet(key)

    # Encrypt the private key
    encrypted_key = fernet.encrypt(data.encode())

    # Store encrypted key and salt
    encrypted_data = {
        "salt": base64.b64encode(salt).decode("utf-8"),
        "encrypted_key": encrypted_key.decode("utf-8"),
    }

    # Create storage directory if it doesn't exist
    storage_dir = Path.home() / ".wallets"
    storage_dir.mkdir(exist_ok=True)

    # Save to file
    with open(storage_dir / filename, "w") as f:
        json.dump(encrypted_data, f)


def load_private_key(password: str, filename: str = "wallet.enc") -> str:
    """
    Load and decrypt an Ethereum private key.

    Args:
        password: Password to decrypt the private key
        filename: Name of file containing encrypted key (default: wallet.enc)

    Returns:
        The decrypted private key
    """
    storage_path = Path.home() / ".wallets" / filename

    # Load encrypted data
    with open(storage_path) as f:
        data = json.load(f)

    salt = base64.b64decode(data["salt"])

    # Recreate key derivation function
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=480000,
    )

    # Regenerate key from password
    key = base64.urlsafe_b64encode(kdf.derive(password.encode()))

    # Decrypt
    fernet = Fernet(key)
    decrypted_key = fernet.decrypt(data["encrypted_key"].encode())

    return decrypted_key.decode()



from asyncio import iscoroutine
from typing import Any

from pydantic.types import Json

from emp_agents.logger import logger


async def execute_tool(
    tools_map,
    function_name: str,
    arguments: Json[Any],
):
    logger.info(f'Executing tool "{function_name}" with arguments {arguments}')

    func = tools_map[function_name]
    response = func(**arguments)
    if iscoroutine(response):
        return await response
    return response



from typing import Any, Awaitable, Callable

from pydantic import BaseModel, Field

from emp_agents.models import GenericTool


class PersistentAgentConfig(BaseModel):
    agent_id: str
    name: str
    description: str
    default_model: str | None = None
    prompt: str = "You are a helpful assistant"
    tools: list[GenericTool | Callable[..., str | Awaitable[str]]] = Field(
        default_factory=list
    )
    requires: list[str] = Field(default_factory=list)
    extra: dict[str, Any] = Field(default_factory=dict)



from pydantic import BaseModel, Field


class Property(BaseModel):
    type: str | None = None
    description: str
    enum: list[str] | None = None
    items: dict[str, "str | Property"] | None = None
    properties: dict[str, "Property"] | None = None
    required: list[str] | None = None


class ToolSchema(BaseModel):
    type: str = "object"
    properties: dict[str, Property]
    required: list[str] = Field(default_factory=list)


class Tool(BaseModel):
    name: str
    description: str
    input_schema: ToolSchema



import json
from enum import StrEnum

from pydantic import BaseModel

from emp_agents.models.shared import AssistantMessage, ToolCall
from emp_agents.types import Role

from .types import AnthropicModelType as ModelType


class ResponseType(StrEnum):
    text = "text"
    tool_use = "tool_use"


class Content(BaseModel):
    id: str | None = None
    type: ResponseType
    text: str | None = None
    input: dict[str, str] | None = None
    name: str | None = None

    @property
    def function(self) -> ToolCall.Function:
        assert self.name is not None
        assert self.input is not None
        return ToolCall.Function(name=self.name, arguments=json.dumps(self.input))

    def to_message(self) -> AssistantMessage:
        return AssistantMessage(content=self.text)


class StopReason(StrEnum):
    tool_use: str = "tool_use"
    end_turn: str = "end_turn"
    max_tokens: str = "max_tokens"
    stop_sequence: str = "stop_sequence"


class Usage(BaseModel):
    input_tokens: int
    output_tokens: int


class Response(BaseModel):
    id: str
    content: list[Content]
    model: ModelType | str
    role: Role
    stop_reason: StopReason
    stop_sequence: str | None
    type: str | None
    usage: Usage

    @property
    def text(self):
        return self.content[0].text

    @property
    def tool_calls(self):
        return [content for content in self.content if content.type == "tool_use"]

    @property
    def messages(self) -> list[AssistantMessage]:
        return [content.to_message() for content in self.content]



from typing import Annotated, Any, Union

from .types import Doc


def is_support_uniontype():
    """Check if current python version supports UnionType or not"""
    try:
        from types import UnionType  # noqa
    except ImportError:
        return False
    return True


def is_doc_meta(
    obj: Annotated[Any, Doc("The object to be checked.")],
) -> Annotated[
    bool, Doc("True if the object is a documentation object, False otherwise.")
]:
    """
    Check if the given object is a documentation object.

    Example:
    >>> is_doc_meta(Doc("This is a documentation object"))
    True
    """
    return getattr(obj, "__class__") == Doc and hasattr(obj, "documentation")


def unwrap_doc(
    obj: Annotated[
        Union[Doc, str], Doc("The object to get the documentation string from.")
    ],
) -> Annotated[str, Doc("The documentation string.")]:
    """
    Get the documentation string from the given object.

    Example:
    >>> unwrap_doc(Doc("This is a documentation object"))
    'This is a documentation object'
    >>> unwrap_doc("This is a documentation string")
    'This is a documentation string'
    """
    if is_doc_meta(obj):
        return obj.documentation
    return str(obj)



from typing import Annotated, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field, PlainSerializer

from emp_agents.models.shared import Message
from emp_agents.models.shared.tools import GenericTool

from .tool import Tool
from .types import OpenAIModelType


class Request(BaseModel):
    """
    https://platform.openai.com/docs/api-reference/chat/create
    """

    model_config = ConfigDict(populate_by_name=True)

    model: OpenAIModelType
    max_tokens: Optional[int] = Field(default=None)
    temperature: Optional[float] = Field(default=None, ge=0, le=2.0)
    tool_choice: Literal["none", "required", "auto", None] = Field(default=None)
    tools: Annotated[
        Optional[list[GenericTool]],
        PlainSerializer(
            lambda tools_list: (
                [tool.to_openai() for tool in tools_list]
                if tools_list is not None
                else None
            ),
            return_type=Optional[list[Tool]],
        ),
    ] = Field(default=None)

    system: str | None = None
    messages: list[Message] | None = None  # Anthropic Field

    frequency_penalty: Optional[float] = Field(default=None, ge=-2.0, le=2.0)  # openai
    presence_penalty: Optional[float] = Field(default=None, ge=-2.0, le=2.0)  # openai
    num_responses: Optional[int] = Field(
        default=None, serialization_alias="n"
    )  # openai
    top_p: Optional[int] = Field(default=None)  # openai

    def model_dump(self, *, exclude_none=True, by_alias=True, **kwargs):
        return super().model_dump(
            exclude_none=exclude_none, by_alias=by_alias, **kwargs
        )



from enum import StrEnum


class AnthropicModelType(StrEnum):
    claude_3_5_sonnet = "claude-3-5-sonnet-20240620"
    claude_3_opus = "claude-3-opus-20240229"
    claude_3_sonnet = "claude-3-sonnet-20240229"
    claude_3_5_haiku = "claude-3-5-haiku-20241022"
    claude_3_haiku = "claude-3-haiku-20240307"
    claude_2_1 = "claude-2.1"
    claude_2_0 = "claude-2.0"
    claude_instant_1_2 = "claude-instant-1.2"



from typing import Literal, TypedDict

try:
    from typing import NotRequired
except ImportError:
    try:
        from typing_extensions import NotRequired
    except ImportError:
        from typing import Optional as NotRequired

try:
    from typing import Doc  # type: ignore
except ImportError:
    try:
        from typing_extensions import Doc  # noqa: F401
    except ImportError:

        class Doc:  # type: ignore
            def __init__(self, documentation: str, /):
                self.documentation = documentation


class ParamSchema(TypedDict):
    """
    Represents the schema for a parameter.
    Attributes:
        type (str): The type of the parameter.
        description (str): The description of the parameter.
        enum (Optional[list[str]]): The list of allowed values for the parameter (optional).
        default (Optional[str]): The default value for the parameter (optional).
    """

    type: str
    description: str
    properties: NotRequired[dict[str, "ParamSchema"]]
    required: NotRequired[list[str]]
    enum: NotRequired[list[str]]
    default: NotRequired[str]


class RootProperty(TypedDict):
    """
    Represents the schema for a parameter.
    Attributes:
        type (str): Root property can only be "object".
        properties (dict[str, ParamSchema]): The properties of the object.
    """

    type: Literal["object"]
    properties: dict[str, ParamSchema]


class FunctionSchema(TypedDict):
    """
    Represents the schema of a function.
    Attributes:
        name (str): The name of the function.
        description (str): The description of the function.
        parameters (RootProperty): The schema for the function parameters.
        input_schema (ParamSchema): The schema for the function parameters if format is "claude".
    """

    name: str
    description: str
    parameters: NotRequired[RootProperty]
    input_schema: NotRequired[RootProperty]



from fast_depends.dependencies.model import Depends


class IgnoreDepends(Depends):
    """
    A Depends that ignores the implicit argument for the tools calls.
    This means the implicit argument must be set before the tools call,
    otherwise the function will fail.
    """



import importlib.util
import pkgutil


def load_tools(directory: str):
    tools = []
    tools_map: dict = {}
    directory_import = directory.rstrip(".").replace("/", ".")
    for importer, pkg_name, is_pkg in pkgutil.iter_modules([directory]):
        if is_pkg:
            module = importlib.import_module(f"{directory_import}.{pkg_name}")
            try:
                tools.extend(module.TOOLS)
            except AttributeError:
                pass
            try:
                tools_map |= module.TOOL_MAP
            except AttributeError:
                pass
    return tools, tools_map



from typing import Annotated

from anthropic.types.message_create_params import ToolChoice, ToolChoiceToolChoiceAuto
from pydantic import BaseModel, ConfigDict, Field, PlainSerializer

from emp_agents.models import Message

from .tool import Tool
from .types import AnthropicModelType


class Request(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    model: AnthropicModelType = Field(default=AnthropicModelType.claude_3_5_sonnet)
    max_tokens: int | None = Field(default=None)
    temperature: float | None = Field(default=None)
    tool_choice: ToolChoice | None = Field(
        default_factory=lambda: ToolChoiceToolChoiceAuto({"type": "auto"})
    )
    tools: list[Tool] = Field(default_factory=list)

    system: str
    messages: Annotated[
        list[Message],
        PlainSerializer(lambda x: [row.serialize_anthropic() for row in x]),
    ]



from typing import ClassVar

from emp_agents.models import GenericTool


class ToolRegistry:
    """Registry to store tool class and method information."""

    _registry: ClassVar[dict[str, type]] = {}

    @classmethod
    def register_class(cls, tool_class: type) -> None:
        """Register a tool class and collect its decorated methods."""
        class_name = tool_class.__name__
        cls._registry[class_name] = tool_class

    @classmethod
    def get_tool(cls, skill_name: str, tool_name: str) -> GenericTool:
        return cls._registry[skill_name]._tools_map[tool_name]

    @classmethod
    def get_skill(cls, skill_name: str) -> list[GenericTool]:
        return cls._registry[skill_name]._tools



from typing import Any, Callable, ParamSpec, TypeVar

from fast_depends import Depends, Provider, inject

from .manager import ImplicitManager
from .models import IgnoreDepends as IgnoreDependsModel

P = ParamSpec("P")
T = TypeVar("T")


def IgnoreDepends(
    dependency: Callable[P, T],
    *,
    use_cache: bool = True,
    cast: bool = True,
) -> Any:
    return IgnoreDependsModel(
        dependency=dependency,
        use_cache=use_cache,
        cast=cast,
    )


def set_implicit(name: str, implicit: Any):
    ImplicitManager.add_implicit(name, implicit)


def lazy_implicit(name: str) -> Callable[P, T] | T:
    return ImplicitManager.lazy_implicit(name)


__all__ = [
    "Depends",
    "ImplicitManager",
    "IgnoreDepends",
    "Provider",
    "inject",
    "set_implicit",
]



from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from emp_agents.models import AssistantMessage, Message, ResponseT, ToolCall

from .types import FinishReason, OpenAIModelType


class Usage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class Choice(BaseModel):
    index: int
    message: AssistantMessage
    logprobs: Optional[str] = None
    finish_reason: FinishReason

    @property
    def content(self):
        return self.message.content


class Response(ResponseT):
    id: str
    object: str
    created: datetime
    model: OpenAIModelType | str
    choices: list[Choice]
    usage: Usage
    system_fingerprint: Optional[str] = None

    @property
    def text(self) -> str:
        return self.choices[0].content

    @property
    def messages(self) -> list[Message]:
        return [self.choices[0].message]

    @property
    def tool_calls(self) -> list[ToolCall] | None:
        return self.choices[0].message.tool_calls

    def __repr__(self):
        return f'<Response id="{self.id}">'

    def print(self):
        for choice in self.choices:
            print(choice.content)
            print("-" * 15)

    __str__ = __repr__



import inspect
import json
import sys
from importlib.util import module_from_spec, spec_from_file_location

from .core import get_function_schema


def print_usage():
    print("Usage: function_schema <file_path> <function_name> [format='openai']")
    print("Example: function_schema ./tests/test_schema.py test_simple_function")
    print("Example: function_schema ./tests/test_schema.py test_simple_function claude")


def main():
    # get cli args of file path
    try:
        file_path = sys.argv[1]
        spec = spec_from_file_location("_defendant", file_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        members = inspect.getmembers(module)

        try:
            format = sys.argv[3]
        except IndexError:
            format = "openai"
        if format not in ["openai", "claude"]:
            print(
                "Invalid format. Use 'openai' or 'claude'. using 'openai'",
                file=sys.stderr,
            )

        for name, func in members:
            if name >= sys.argv[2]:
                print(json.dumps(get_function_schema(func, format), indent=2))
                sys.exit(0)
        print(f"Function {sys.argv[2]} not found in {file_path}")
    except IndexError:
        print_usage()
        sys.exit(1)


if __name__ == "__main__":
    main()



import pytest

from emp_agents.agents.persistentagent import PersistentAgent, PersistentAgentConfig
from emp_agents.providers.openai import OpenAIProvider


@pytest.mark.asyncio(scope="session")
async def test_from_config():
    config = PersistentAgentConfig(
        agent_id="test-agent-1",
        name="Test Agent",
        description="A test agent",
        default_model="gpt-3.5-turbo",
        prompt="You are a test assistant",
        tools=[],
        requires=[],
    )
    agent = PersistentAgent.from_config(config, provider=OpenAIProvider())

    assert isinstance(agent, PersistentAgent)
    assert agent.config == config
    assert agent.description == config.description
    assert agent.default_model == config.default_model
    assert agent.prompt == config.prompt
    assert agent.tools == config.tools
    assert agent.requires == config.requires


@pytest.mark.asyncio(scope="session")
async def test_perform_action(capsys):
    config = PersistentAgentConfig(
        agent_id="test-agent-1",
        name="Test Agent",
        description="A test agent",
        default_model="gpt-3.5-turbo",
        prompt="You are a test assistant",
        tools=[],
        requires=[],
    )
    agent = PersistentAgent.from_config(config, provider=OpenAIProvider())
    agent.perform_action()

    captured = capsys.readouterr()
    expected_output = (
        f"Agent {config.name} (ID: {config.agent_id}) is performing an action.\n"
    )
    assert captured.out == expected_output



from datetime import datetime, timezone
from typing import Annotated

from eth_rpc import PrivateKeyWallet
from eth_rpc.networks import Network, get_network_by_name
from eth_typeshed.erc20 import ERC20
from eth_typeshed.uniswap_v2.router.contract import (
    EthSwapRequest,
    TokenSwapRequest,
    UniswapV2Router,
)
from eth_typing import HexAddress, HexStr
from typing_extensions import Doc

from .constants import ROUTER_ADDRESSES


async def swap_exact_tokens_for_tokens(
    token_in: Annotated[HexAddress, Doc("The token to swap from")],
    token_out: Annotated[HexAddress, Doc("The token to swap to")],
    amount_in: Annotated[float, Doc("The amount of tokens to swap")],
    recipient: Annotated[HexAddress, Doc("The recipient of the swapped tokens")],
    slippage: Annotated[float, Doc("The slippage tolerance")],
    deadline: Annotated[int | None, Doc("The deadline for the swap")],
    network: type[Network],
    wallet: PrivateKeyWallet,
) -> HexStr:
    if not deadline:
        deadline = int(datetime.now(timezone.utc).timestamp()) + 60
    assert deadline
    network_type = get_network_by_name(network)
    token = ERC20[network_type](address=token_in)
    decimals = await token.decimals().get()
    amount_in_wei = amount_in * 10**decimals
    router_address = ROUTER_ADDRESSES[network]
    router = UniswapV2Router[network_type](address=router_address)
    amount_out = await router.get_amounts_out(
        amount_in=amount_in_wei,
        path=[token_in, token_out],
    ).get()
    amount_out_min = amount_out[-1] * (1 - slippage)
    return await router.swap_exact_tokens_for_tokens_supporting_fee_on_transfer_tokens(
        TokenSwapRequest(
            amount_in=amount_in_wei,
            amount_out_min=amount_out_min,
            route=[
                token_in,
                token_out,
            ],
            to=recipient,
            deadline=deadline,
        )
    ).execute(wallet)


async def swap_exact_eth_for_tokens(
    token_out: HexAddress,
    amount_in: float,
    recipient: HexAddress,
    slippage: float,
    deadline: Annotated[int | None, Doc("The deadline for the swap")],
    network: type[Network],
    wallet: PrivateKeyWallet,
) -> HexStr:
    if not deadline:
        deadline = int(datetime.now(timezone.utc).timestamp()) + 60
    assert deadline
    decimals = 18
    amount_in_wei = amount_in * 10**decimals
    router_address = ROUTER_ADDRESSES[network]
    router = UniswapV2Router[network](address=router_address)
    weth_address = await router.weth().get()
    amount_out = await router.get_amounts_out(
        amount_in=amount_in_wei,
        path=[weth_address, token_out],
    ).get()
    amount_out_min = amount_out[-1] * (1 - slippage)
    return await router.swap_exact_eth_for_tokens_supporting_fee_on_transfer_tokens(
        EthSwapRequest(
            amount_out_min=amount_out_min,
            route=[
                weth_address,
                token_out,
            ],
            to=recipient,
            deadline=deadline,
        )
    ).execute(wallet, value=amount_in_wei)


async def swap_exact_tokens_for_eth(
    token_in: Annotated[HexAddress, Doc("The token to swap from")],
    amount_in: Annotated[float, Doc("The amount of tokens to swap")],
    recipient: Annotated[HexAddress, Doc("The recipient of the swapped tokens")],
    slippage: Annotated[float, Doc("The slippage tolerance")],
    deadline: Annotated[int | None, Doc("The deadline for the swap")],
    network: type[Network],
    wallet: PrivateKeyWallet,
) -> HexStr:
    if not deadline:
        deadline = int(datetime.now(timezone.utc).timestamp()) + 60
    assert deadline
    token = ERC20[network](address=token_in)
    decimals = await token.decimals().get()
    amount_in_wei = amount_in * 10**decimals
    router_address = ROUTER_ADDRESSES[network]
    router = UniswapV2Router[network](address=router_address)
    weth_address = await router.weth().get()
    amount_out = await router.get_amounts_out(
        amount_in=amount_in_wei,
        path=[token_in, weth_address],
    ).get()
    amount_out_min = amount_out[-1] * (1 - slippage)
    return await router.swap_exact_tokens_for_eth_supporting_fee_on_transfer_tokens(
        TokenSwapRequest(
            amount_in=amount_in_wei,
            amount_out_min=amount_out_min,
            route=[
                token_in,
                weth_address,
            ],
            to=recipient,
            deadline=deadline,
        )
    ).execute(wallet)



from typing import Any, Callable

from fast_depends import Provider
from pydantic import BaseModel, ConfigDict, Field

from emp_agents.models import Message
from emp_agents.models.protocol import SkillSet

from .base import AgentBase


class SkillsAgent(AgentBase):
    skills: list[type[SkillSet]] = Field(default_factory=list)
    scopes: list[tuple[Provider, Callable, Callable]] = Field(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, __context: Any):
        super().model_post_init(__context)

        for skill in self.skills:
            for tool in skill._tools:
                self._add_tool(tool)

    async def _run_conversation(
        self,
        messages: list[Message],
        model: str,
        max_tokens: int | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        for scope, old, new in self.scopes:
            scope.dependency_overrides[old] = new

        response = await super()._run_conversation(
            messages, model, max_tokens, response_format
        )

        for scope, old, new in self.scopes:
            scope.dependency_overrides.pop(old, None)

        return response

    async def respond(
        self,
        question: str,
        model: str | None = None,
        max_tokens: int | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        for scope, old, new in self.scopes:
            scope.dependency_overrides[old] = new

        response = await super().respond(question, model, max_tokens, response_format)

        for scope, old, new in self.scopes:
            scope.dependency_overrides.pop(old, None)

        return response



from emp_agents.agents.base import AgentBase
from emp_agents.config.agent_config import PersistentAgentConfig
from emp_agents.models import Provider


class PersistentAgent(AgentBase):
    config: PersistentAgentConfig

    @classmethod
    def from_config(
        cls, config: PersistentAgentConfig, provider: Provider
    ) -> "PersistentAgent":
        return cls(
            config=config,
            provider=provider,
            description=config.description,
            default_model=config.default_model,
            prompt=config.prompt,
            tools=config.tools,
            requires=config.requires,
            **config.extra,
        )

    def perform_action(self):
        print(
            f"Agent {self.config.name} (ID: {self.config.agent_id}) is performing an action."
        )



from typing import Annotated, Literal, Optional, cast

from eth_rpc.networks import Network, get_network_by_name
from typing_extensions import Doc

from emp_agents.models.protocol import SkillSet, tool_method

_network: Optional[type[Network]] = None


NetworkOptions = Literal[
    "ethereum",
    "sepolia",
    "base",
    "BaseSepolia",
    "arbitrum",
    "ArbitrumSepolia",
]


class NetworkSkill(SkillSet):
    """
    Tools for controlling the current network
    """

    @tool_method
    @staticmethod
    def set_network(
        network: Annotated[
            NetworkOptions,
            Doc("The network to set"),
        ],
    ):
        """
        Set the current network
        """
        global _network
        try:
            _network = get_network_by_name(network)
        except ValueError:
            return f"Invalid network: {network}"
        return f"network set to {network}"

    @tool_method
    @staticmethod
    def get_network() -> str:
        """
        Get the current network
        """
        if not _network:
            return "No network set, try setting the network first"
        return f"current network: {_network.name}"

    @tool_method
    @staticmethod
    def make_block_explorer_link(tx_hash: str) -> str:
        if not _network:
            return "No network set, try setting the network first"
        return f"{_network.block_explorer.url}/tx/{tx_hash}"

    @staticmethod
    def get_network_type() -> type[Network] | None:
        return _network

    @staticmethod
    def get_network_str() -> NetworkOptions | None:
        if not _network:
            return None
        return cast(NetworkOptions, _network.name)



import os
from typing import ClassVar

from anthropic import AsyncAnthropic as Anthropic
from anthropic import NotFoundError
from pydantic import ConfigDict, Field, PrivateAttr

from emp_agents.exceptions import InvalidModelException
from emp_agents.models import Provider, Request, Role

from .response import Response
from .types import AnthropicModelType


class AnthropicProvider(Provider[Response]):
    URL: ClassVar[str] = "https://api.openai.com/v1/chat/completions"

    api_key: str = Field(default_factory=lambda: os.environ["ANTHROPIC_API_KEY"])
    default_model: AnthropicModelType = Field(
        default=AnthropicModelType.claude_3_5_sonnet
    )
    _client: Anthropic = PrivateAttr()

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, __context) -> None:
        self._client = Anthropic(api_key=self.api_key)
        return super().model_post_init(__context)

    @property
    def headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }

    async def completion(self, request: Request) -> Response:
        try:
            message = await self._client.messages.create(**self._from_request(request))
            return Response(**message.model_dump())
        except NotFoundError as e:
            raise InvalidModelException(e)

    def _from_request(self, request: Request):
        exclude = ["frequency_penalty", "presence_penalty", "num_responses", "n"]
        result = request.model_dump(exclude_none=True)
        result["tools"] = (
            [t.to_anthropic() for t in request.tools] if request.tools else []
        )
        if "tool_choice" in result:
            result["tool_choice"] = {"type": result["tool_choice"]}
        for field in exclude:
            if field in result:
                del result[field]
        messages = [
            message for message in request.messages if message.role != Role.system
        ]
        system_messages = [
            message for message in request.messages if message.role == Role.system
        ]
        result["system"] = result.get("system", "") + str(
            "\n".join([m.content for m in system_messages])
        )
        result["messages"] = [m.model_dump(exclude_none=True) for m in messages]

        if "response_format" in result:
            result[
                "system"
            ] += f"""
            Always give your response in JSON format, with no additional text.  This supersedes any other guidance.
            Make sure it is a valid JSON, matching this format.  Dont include any additional text or decorations:
            ```json
            {result["response_format"].model_json_schema()}
            ```
            """
            del result["response_format"]
        return result



from emp_agents.models.protocol.decorators import (
    cachable,
    onchain_action,
    tool_method,
    view_action,
)
from emp_agents.models.protocol.skill_set import SkillSet

__all__ = ["SkillSet", "cachable", "tool_method", "onchain_action", "view_action"]



from enum import StrEnum

from pydantic import BaseModel


class Classification(BaseModel):
    name: str
    description: str


class FinishReason(StrEnum):
    stop = "stop"
    length = "length"
    function_call = "function_call"
    content_filter = "content_filter"
    null = "null"
    tool_calls = "tool_calls"


class OpenAIModelType(StrEnum):
    gpt3_5 = "gpt-3.5-turbo-0125"
    gpt3_5_turbo = "gpt-3.5-turbo"
    gpt4 = "gpt-4"
    gpt4_turbo = "gpt-4-turbo"
    gpt4o_mini = "gpt-4o-mini"  # 128_000 tokens
    gpt4o = "gpt-4o"
    gpt_o1_mini = "o1-mini"
    gpt_o1_preview = "o1-preview"
    o1 = "o1"
    o1_24_12_17 = "o1-2024-12-17"



from contextvars import ContextVar
from typing import Annotated, Optional

from eth_rpc import Account, PrivateKeyWallet
from eth_rpc.networks import Network, get_network_by_name
from eth_typing import HexAddress, HexStr
from typing_extensions import Doc

from emp_agents.implicits import Depends, inject
from emp_agents.models.protocol import SkillSet, onchain_action, tool_method

from ..network import NetworkOptions, NetworkSkill

# Context variable to store private key
_private_key: ContextVar[Optional[str]] = ContextVar("_private_key", default=None)


class SimpleWalletSkill(SkillSet):
    """A simple wallet tool that stores private key in memory using context vars"""

    @staticmethod
    def get_wallet() -> PrivateKeyWallet:
        key = _private_key.get()
        if key is None:
            raise ValueError("No private key set")
        return PrivateKeyWallet(private_key=HexStr(key))

    @tool_method
    @staticmethod
    def create_wallet() -> str:
        """Create a new private key wallet"""

        wallet = PrivateKeyWallet.create_new()
        _private_key.set(wallet.private_key)
        return (
            f"Wallet created: {wallet.address} with private key: {wallet.private_key}"
        )

    @tool_method
    @staticmethod
    def set_private_key(
        private_key: Annotated[str, Doc("The private key to set")]
    ) -> str:
        """Set the private key in the context"""

        _private_key.set(private_key)
        return "Private key set successfully"

    @tool_method
    @staticmethod
    def get_private_key() -> str:
        """Get the private key from the context"""

        key = _private_key.get()
        if key is None:
            return "No private key set"
        return key

    @tool_method
    @staticmethod
    def clear_private_key() -> str:
        """Clear the private key from the context"""

        _private_key.set(None)
        return "Private key cleared"

    @tool_method
    @staticmethod
    def get_address() -> str:
        """Get the address of the wallet"""

        key = _private_key.get()
        if key is None:
            return "No private key set"
        wallet = PrivateKeyWallet(private_key=HexStr(key))
        return wallet.address

    @tool_method
    @staticmethod
    @inject
    async def get_eth_balance(
        address: Annotated[
            HexAddress, Doc("The address to get the balance of")
        ] = Depends(get_address),
        network: Annotated[
            NetworkOptions | None,
            Doc("The network to use"),
        ] = Depends(NetworkSkill.get_network_str),
    ) -> str:
        """Get the balance of the wallet in ETH"""

        if network is None:
            return "No network set, try setting the network first"

        network_type: type[Network] = get_network_by_name(network)
        balance = await Account[network_type].get_balance(address)  # type: ignore[valid-type]
        return f"Balance: {balance}"

    @onchain_action
    @staticmethod
    @inject
    async def transfer_eth(
        recipient: Annotated[HexAddress, Doc("The address to transfer the ETH to")],
        amount: Annotated[int, Doc("The amount of ETH to transfer, in wei")],
        message: Annotated[
            str | None,
            Doc(
                "Any data you want to include in the transfer.  It will be utf-8 encoded"
            ),
        ] = None,
        network: Annotated[
            NetworkOptions | None,
            Doc("The network to use"),
        ] = Depends(NetworkSkill.get_network_str),
    ) -> str:
        """Transfer ETH to a recipient"""

        if network is None:
            return "No network set, try setting the network first"
        wallet = SimpleWalletSkill.get_wallet()
        if wallet is None:
            return "No wallet set, try creating a wallet first"

        network_type: type[Network] = get_network_by_name(network)
        balance = await Account[network_type].get_balance(wallet.address)  # type: ignore[valid-type]
        if message:
            data = "0x" + message.encode("utf-8").hex()
        else:
            data = "0x"

        if balance <= amount:
            return "Insufficient balance"

        tx = await wallet[network_type].transfer(recipient, amount, data=data)
        return f"Transaction sent: {tx}"



from typing import Annotated, Literal

from eth_rpc.networks import get_network_by_name
from eth_typeshed.uniswap_v2.factory import GetPairRequest, UniswapV2Factory
from eth_typeshed.uniswap_v2.pair import UniswapV2Pair
from eth_typing import HexAddress
from typing_extensions import Doc

from .constants import FACTORY_ADDRESSES


async def get_price(
    network: Annotated[
        Literal["ethereum", "arbitrum", "base"],
        Doc("The network to query for price"),
    ],
    token_in: Annotated[HexAddress, Doc("The token to swap from")],
    token_out: Annotated[HexAddress, Doc("The token to swap to")],
) -> float:
    """Calculate the price of a token in terms of another token on UniswapV2"""

    factory_address = FACTORY_ADDRESSES[network]

    network_type = get_network_by_name(network)
    factory = UniswapV2Factory[network_type](address=factory_address)
    pair = await factory.get_pair(
        GetPairRequest(
            token_a=token_in,
            token_b=token_out,
        )
    ).get()
    return await UniswapV2Pair[network_type](address=pair.address).get_price(
        token0=True
    )



from collections.abc import Callable
from typing import Any


class ImplicitManager:
    implicits: dict[str, Any] = {}

    @classmethod
    def get_implicit(cls, name: str):
        """Get an implicit argument by name"""
        if name not in cls.implicits:
            raise ValueError(f"Implicit {name} not found")
        return cls.implicits[name]

    @classmethod
    def add_implicit(cls, name: str, implicit: Any):
        """Adds a new implicit argument"""
        cls.implicits[name] = implicit

    @classmethod
    def lazy_implicit(cls, name: str) -> Callable[..., Any] | Any:
        """Create a lazy implicit that can be resolved later"""

        def _get_implicit(*args, **kwargs) -> Any:
            imp = cls.get_implicit(name)
            if callable(imp):
                return imp(*args, **kwargs)
            return imp

        return _get_implicit



import json
from typing import Annotated, Literal

import httpx
from typing_extensions import Doc

from emp_agents.logger import logger
from emp_agents.models.protocol import SkillSet, view_action


class GmxSkill(SkillSet):
    @view_action
    @staticmethod
    async def get_tokens_address_dict(
        chain: Annotated[
            Literal["arbitrum", "avalanche"],
            Doc("The chain to query for tokens"),
        ] = "arbitrum",
    ):
        """
        Query the GMX infra api for to generate dictionary of tokens available on v2.  Returns
        a JSON of tokens available on the given chain.
        """

        url = {
            "arbitrum": "https://arbitrum-api.gmxinfra.io/tokens",
            "avalanche": "https://avalanche-api.gmxinfra.io/tokens",
        }

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(url[chain])

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Parse the JSON response
                token_infos = response.json()["tokens"]
            else:
                logger.error(f"Error: {response.status_code}")
        except Exception as e:
            print(f"Error: {e}")

        token_address_dict = {}

        for token_info in token_infos:
            token_address_dict[token_info["address"]] = token_info

        return json.dumps(token_address_dict)



from asyncio import iscoroutinefunction

from emp_agents.exceptions import TooManyTriesException


def retry(times):
    def func_wrapper(f):
        async def wrapper(*args, **kwargs):
            for _ in range(times):
                try:
                    if iscoroutinefunction(f):
                        return await f(*args, **kwargs)
                    return f(*args, **kwargs)
                except Exception as _exc:
                    exc = _exc
            raise TooManyTriesException() from exc

        return wrapper

    return func_wrapper



from pydantic import BaseModel, Field


class Property(BaseModel):
    type: str | None = None
    description: str
    enum: list[str] | None = None
    items: dict[str, "str | Property"] | None = None
    properties: dict[str, "Property"] | None = None
    required: list[str] | None = None


class Parameters(BaseModel):
    type: str = Field(default="object")
    properties: dict[str, Property]
    required: list[str] | None = None

    @classmethod
    def build(cls, type: str = "object", **kwargs):
        return cls(type=type, **kwargs)


class Function(BaseModel):
    description: str
    name: str
    parameters: Parameters | None
    # required: list[str] | None = Field(default=None)
    additional_properties: bool = False


class Tool(BaseModel):
    type: str = Field(default="function")
    function: Function
    strict: bool = True

    @classmethod
    def build(cls, type="function", **kwargs):
        return cls(type=type, **kwargs)



import pytest

from emp_agents.agents.planner import Planner, Task, TaskList
from emp_agents.providers.openai import OpenAIProvider


@pytest.mark.asyncio(scope="session")
async def test_generate():
    agent = Planner(
        provider=OpenAIProvider(),
    )
    subject = "write a blog post"
    task_list = await agent.generate(subject)
    assert isinstance(task_list, TaskList)

    assert len(task_list.tasks) > 0

    for task in task_list.tasks:
        assert isinstance(task, Task)
        assert isinstance(task.title, str)
        assert isinstance(task.description, str)
        assert task.title != ""
        assert task.description != ""



from emp_agents.models.shared.message import AssistantMessage, ToolMessage, UserMessage

QuestionAnswer = tuple[UserMessage, AssistantMessage]
TooledMessageSequence = tuple[UserMessage, AssistantMessage, list[ToolMessage]]
AgenticMessageSequence = tuple[
    UserMessage,
    list[tuple[AssistantMessage, list[ToolMessage]]],
    AssistantMessage,
]

Example = QuestionAnswer | TooledMessageSequence | AgenticMessageSequence



import json
from typing import Annotated, Callable, Literal

from eth_rpc import PrivateKeyWallet
from eth_rpc.networks import Network
from eth_typing import HexAddress
from typing_extensions import Doc

from emp_agents.implicits import IgnoreDepends, Provider, inject
from emp_agents.models.protocol import SkillSet, onchain_action, view_action

from ..network import NetworkSkill
from ..wallets import SimpleWalletSkill
from .price import get_price
from .swap import (
    swap_exact_eth_for_tokens,
    swap_exact_tokens_for_eth,
    swap_exact_tokens_for_tokens,
)

uniswap_scope = Provider()


def load_wallet() -> PrivateKeyWallet | None:
    """
    This can be overridden by using the scope for your agent.
    Use `scope_load_wallet` to scope this method.
    """
    return SimpleWalletSkill.get_wallet()


def load_network() -> type[Network] | None:
    """
    This can be overridden by using the scope for your agent.
    Use `scope_load_network` to scope this method.
    """
    return NetworkSkill.get_network_type()


def scope_load_wallet(
    new_load_wallet: Callable[..., PrivateKeyWallet]
) -> tuple[Provider, Callable, Callable]:
    return (uniswap_scope, load_wallet, new_load_wallet)


class UniswapSkill(SkillSet):
    """
    Skills for interacting with UniswapV2.
    """

    @view_action
    @staticmethod
    @inject(dependency_overrides_provider=uniswap_scope)
    async def get_price(
        network: Annotated[
            Literal["ethereum", "arbitrum", "base"],
            Doc("The network to query for price"),
        ],
        token_in: Annotated[HexAddress, Doc("The token to swap from")],
        token_out: Annotated[HexAddress, Doc("The token to swap to")],
    ) -> str:
        """Get the price of a token in terms of another token on UniswapV2"""

        price = await get_price(network, token_in, token_out)
        return json.dumps({"price": str(price)})

    @onchain_action
    @staticmethod
    @inject(dependency_overrides_provider=uniswap_scope)  # type: ignore[call-overload]
    async def swap(
        input_token: Annotated[
            HexAddress | None, Doc("The token to swap from.  None if ETH")
        ],
        output_token: Annotated[
            HexAddress | None, Doc("The token to swap to. None if ETH")
        ],
        amount_in: Annotated[float, Doc("The amount of tokens to swap")],
        recipient: Annotated[HexAddress, Doc("The recipient of the swapped tokens")],
        slippage: Annotated[
            float, Doc("The slippage tolerance, defaults to 0.5%")
        ] = 0.005,
        deadline: Annotated[
            int | None,
            Doc("The deadline for the swap, defaults to 1 minute from now"),
        ] = None,
        network: type[Network] | None = IgnoreDepends(load_network),
        wallet: PrivateKeyWallet | None = IgnoreDepends(load_wallet),
    ) -> str:
        """Swap an exact amount of tokens for ETH.  Returns the transaction hash or error message."""

        if wallet is None:
            return "WalletNotFoundError: make sure you have setup your wallet loading logic"

        if network is None:
            return "NetworkNotFoundError: make sure you have setup your network loading logic"

        if input_token is None:
            if output_token is None:
                return "InvalidSwapError: both input and output tokens cannot be None"
            assert output_token is not None
            return await swap_exact_eth_for_tokens(
                output_token,
                amount_in,
                recipient,
                slippage,
                deadline,
                network,
                wallet,
            )
        elif output_token is None:
            return await swap_exact_tokens_for_eth(
                input_token,
                amount_in,
                recipient,
                slippage,
                deadline,
                network,
                wallet,
            )
        return await swap_exact_tokens_for_tokens(
            input_token,
            output_token,
            amount_in,
            recipient,
            slippage,
            deadline,
            network,
            wallet,
        )



from typing import Awaitable, Callable

StrCallable = Callable[..., str | Awaitable[str]]


def add_to_decorated_functions(func: Callable, decorator_name: str):
    """Helper function to add function metadata to the function object."""
    tool_name = func.__name__
    tool_description = func.__doc__ or ""

    if tool_name and tool_description:
        if not hasattr(func, "_tool_metadata"):
            func._tool_metadata = {}

        func._tool_metadata.update(
            {
                "name": tool_name,
                "description": tool_description,
                "decorator": decorator_name,
            }
        )


def tool_method(func: StrCallable):
    """Decorator that marks a method as a protocol tool method.

    This allows the method to be exposed as a tool that can be used by AI agents
    to interact with protocols. The method will be added to the class's tools list
    when the class is created.

    Args:
        func: The method to mark as a protocol tool

    Returns:
        The decorated method
    """
    # Set attribute to mark this as a protocol tool method
    method = func
    setattr(method, "_is_tool_method", True)
    add_to_decorated_functions(method, "tool_method")
    return method


def onchain_action(func: StrCallable):
    """Decorator that marks a method as an onchain action.

    This indicates that the method will submit transactions or otherwise modify blockchain state.
    The method will be added to the class's tools list when the class is created.

    Args:
        func: The method to mark as an onchain action

    Returns:
        The decorated method
    """
    # Set attributes to mark this as a protocol tool method and onchain action
    method = func
    setattr(method, "_is_tool_method", True)
    setattr(method, "_is_onchain_action", True)
    add_to_decorated_functions(method, "onchain_action")
    return method


def view_action(func: StrCallable):
    """Decorator that marks a method as a view-only data fetch action.

    This indicates that the method will only read blockchain state without modifying it.
    The method will be added to the class's tools list when the class is created.

    Args:
        func: The method to mark as a view action

    Returns:
        The decorated method
    """
    # Set attributes to mark this as a protocol tool method and view action
    method = func
    setattr(method, "_is_tool_method", True)
    setattr(method, "_is_view_action", True)
    add_to_decorated_functions(method, "view_action")
    return method


def cachable(func: StrCallable):
    """Decorator that caches the results of a method."""
    cache = {}

    def wrapper(*args, **kwargs):
        # Create cache key from arguments
        key = str(args) + str(sorted(kwargs.items()))

        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]

    # Preserve the original function attributes
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__

    # Set attribute to mark this as a protocol tool method
    setattr(wrapper, "_is_tool_method", True)
    add_to_decorated_functions(wrapper, "cachable")
    return wrapper



import asyncio
from textwrap import dedent
from typing import Any, Awaitable, Callable

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    computed_field,
    field_validator,
)

from emp_agents.agents.history import AbstractConversationProvider, ConversationProvider
from emp_agents.logger import logger
from emp_agents.models import (
    AssistantMessage,
    GenericTool,
    Message,
    Provider,
    Request,
    ResponseT,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from emp_agents.models.middleware import Middleware
from emp_agents.providers.openai import OpenAIModelType
from emp_agents.types import Role
from emp_agents.utils import count_tokens, execute_tool, summarize_conversation


class AgentBase(BaseModel):
    agent_id: str = Field(default="")
    description: str = Field(default="")
    default_model: str | None = None
    prompt: str = Field(default="You are a helpful assistant")
    tools: list[GenericTool | Callable[..., str | Awaitable[str]]] = Field(
        default_factory=list
    )
    requires: list[str] = []
    provider: Provider
    conversation: AbstractConversationProvider = Field(
        default_factory=ConversationProvider
    )

    # This can be used to modify the conversation before completion, such as RAG
    middleware: list[Middleware] = Field(default_factory=list)

    _tools: list[GenericTool] = PrivateAttr(default_factory=list)
    _tools_map: dict[str, Callable[..., Any]] = PrivateAttr(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def conversation_history(self) -> list[Message]:
        return self.conversation.get_history()

    @computed_field  # type: ignore[prop-decorator]
    @property
    def _default_model(self) -> str:
        if self.default_model:
            return self.default_model
        _model = self.provider.default_model
        assert _model is not None, "Default model is required"
        return _model

    def _load_model(self, model: str | None) -> str:
        if model is None:
            model = self._default_model
        assert model is not None, "Model is required"
        return model

    @field_validator("prompt", mode="before")
    @classmethod
    def to_prompt(cls, v: str) -> str:
        return dedent(v).strip()

    @field_validator("tools", mode="before")
    @classmethod
    def to_generic_tools(
        cls, v: list[Callable[..., Any] | GenericTool]
    ) -> list[GenericTool]:
        return [
            GenericTool.from_func(tool) if not isinstance(tool, GenericTool) else tool
            for tool in v
        ]

    def _load_implicits(self):
        """Override this method to load implicits to the agent directly"""

    def model_post_init(self, _context: Any):
        if not (self.provider.api_key):
            raise ValueError("Must provide an api key")

        for tool in self.tools:
            if isinstance(tool, GenericTool):
                self._tools.append(tool)
            else:
                self._tools.append(GenericTool.from_func(tool))

        self._tools_map = {tool.name: tool.func for tool in self._tools}
        self.conversation.add_message(SystemMessage(content=self.system_prompt))

        self._load_implicits()

    def get_token_count(
        self,
        model: str | OpenAIModelType = "gpt4o_mini",
    ) -> int:
        """A utility to get the token count for openai models, fairly accurate across all providers"""
        return count_tokens(self.conversation.get_history(), model)

    async def summarize(
        self,
        model: str | None = None,
        update: bool = True,
        prompt: str | None = None,
        max_tokens: int = 500,
    ) -> str:
        model = self._load_model(model)

        summary = await summarize_conversation(
            self.provider,
            self.conversation.get_history(),
            model=model,
            prompt=prompt,
            max_tokens=max_tokens,
        )
        if update:
            self.conversation.set_history([summary])
        assert summary.content is not None, "Summary content should always be present"
        return summary.content

    async def respond(
        self,
        question: str,
        model: str | None = None,
        max_tokens: int | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        """Send a one-off question and get a response"""
        if model is None:
            model = self._default_model

        conversation = [
            SystemMessage(content=self.system_prompt),
            UserMessage(content=question),
        ]
        return await self._run_conversation(
            conversation,
            model=model,
            max_tokens=max_tokens,
            response_format=response_format,
        )

    async def complete(
        self,
        model: str | None = None,
        max_tokens: int | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        """Complete the current conversation until no more tool calls"""
        _model = self._load_model(model)
        return await self._run_conversation(
            self.conversation.get_history(),
            model=_model,
            max_tokens=max_tokens,
            response_format=response_format,
        )

    async def _run_conversation(
        self,
        messages: list[Message],
        model: str,
        max_tokens: int | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        """Core conversation loop handling tool calls"""
        conversation = messages.copy()
        for middleware in self.middleware:
            _conversation = middleware.process(conversation)
            if isinstance(_conversation, Awaitable):
                conversation = await _conversation  # type: ignore
            else:
                conversation = _conversation
        while True:
            request = Request(
                messages=conversation,
                model=model,
                tools=self._tools,
                max_tokens=max_tokens or 1_000,
                response_format=response_format,
            )
            response: ResponseT = await self.provider.completion(request)
            conversation += response.messages

            if not response.tool_calls:
                self.conversation.set_history(conversation)
                return response.text

            tool_invocation_coros = [
                execute_tool(
                    self._tools_map,
                    tool_call.function.name,
                    tool_call.function.arguments,
                )
                for tool_call in response.tool_calls
            ]
            tool_results = await asyncio.gather(*tool_invocation_coros)
            for result, tool_call in zip(tool_results, response.tool_calls):
                message = ToolMessage(
                    content=result,
                    tool_call_id=(
                        tool_call.id if tool_call and hasattr(tool_call, "id") else None
                    ),
                )
                if hasattr(self, "conversation_history"):
                    logger.info(message)
                conversation += [message]
                self.conversation.set_history(conversation)

    async def answer(
        self,
        question: str,
        model: str | None = None,
        response_format: type[BaseModel] | None = None,
    ) -> str:
        self.conversation.add_message(Message(role=Role.user, content=question))

        return await self.complete(
            model=model,
            response_format=response_format,
        )

    def add_message(
        self,
        message: Message,
    ) -> None:
        self.conversation.add_message(message)

    def add_messages(
        self,
        messages: list[Message],
    ) -> None:
        self.conversation.add_messages(messages)

    async def __call__(
        self,
        question: str,
        model: str | None = None,
    ) -> str:
        return await self.answer(question, model=model)

    async def reset(self):
        self.conversation.reset()

    @property
    def system_prompt(self) -> str:
        prompt = self.prompt
        return prompt.strip()

    def print_conversation(self) -> None:
        for message in self.conversation.get_history():
            print(f"{message.role}: {message.content}")

    def _make_message(self, content: str, role: Role = Role.user) -> Message:
        return Message.build(content, role)

    async def run(self):
        conversation = [SystemMessage(content=self.system_prompt)]
        while True:
            question = input("You: ")
            if question == "":
                break
            conversation += [UserMessage(content=question)]
            response = await self.answer(question)
            print(response)
            conversation += [AssistantMessage(content=response)]

    def _add_tool(self, tool: GenericTool) -> None:
        self._tools.append(tool)
        self._tools_map[tool.name] = tool.func

    def run_sync(self):
        asyncio.run(self.run())

    def __repr__(self):
        prompt = self.prompt[:100].strip().replace("\n", " ")
        if len(prompt) >= 50:
            prompt = prompt[:50] + "..."
        return dedent(
            """
            <{class_name}
                prompt="{prompt}..."
                tools=[
                    {tools}
                ]
            >
        """.format(
                class_name=self.__class__.__name__,
                prompt=prompt,
                tools="\n".join([repr(tool) for tool in self.tools]),
            )
        ).strip()

    __str__ = __repr__



from abc import abstractmethod
from typing import TypeVar

from pydantic import BaseModel

from ..types import TCompletionAgent
from .shared import Message, Request, ToolCall


class ResponseT(BaseModel):
    @property
    @abstractmethod
    def text(self) -> str:
        """The text from the API response"""

    @property
    @abstractmethod
    def messages(self) -> list[Message]:
        """The API response messages"""

    @property
    @abstractmethod
    def tool_calls(self) -> list[ToolCall]: ...


Response = TypeVar("Response", bound=ResponseT)


class Provider(BaseModel, TCompletionAgent[Response]):
    api_key: str | None = None
    default_model: str | None = None

    def _load_model(self, model: str | None) -> str:
        if model is None:
            model = self.default_model
        if not model:
            raise ValueError("No model provided")
        return model

    @abstractmethod
    async def completion(self, request: Request) -> Response: ...



from emp_agents.models.shared.example import Example
from emp_agents.models.shared.message import (
    AssistantMessage,
    Message,
    SystemMessage,
    ToolCall,
    ToolMessage,
    UserMessage,
)
from emp_agents.models.shared.request import Request
from emp_agents.models.shared.tools import GenericTool, Property
from emp_agents.types.enums import Role

__all__ = [
    "GenericTool",
    "Message",
    "Property",
    "Request",
    "Role",
    "SystemMessage",
    "ToolCall",
    "ToolMessage",
    "UserMessage",
    "AssistantMessage",
    "Example",
]



from typing import Annotated, Literal

from typing_extensions import Doc

from emp_agents.models.protocol import SkillSet, tool_method
from emp_agents.tools.dexscreener.api import DexScreenerApi


class DexScreenerSkill(SkillSet):
    """
    Skill for interacting with DexScreener API
    Documentation: https://docs.dexscreener.com/api/reference
    """

    @tool_method
    @staticmethod
    async def search_pairs(
        query: Annotated[str, Doc("The search query to find trading pairs")],
    ) -> str:
        """Search for pairs matching query (rate-limit 300 requests per minute)"""
        return await DexScreenerApi.search_pairs(query)

    @tool_method
    @staticmethod
    async def get_pair_by_chain(
        chain_id: Annotated[
            Literal["ethereum", "solana", "arbitrum", "base", "bsc"],
            Doc("The chain to search on"),
        ],
        pair_id: Annotated[str, Doc("The pair contract address")],
    ) -> str:
        """Retrieve trading pairs by specified chain and pair address (rate-limit 300 requests per minute)"""
        return await DexScreenerApi.get_pair_by_chain(chain_id, pair_id)

    @tool_method
    @staticmethod
    async def find_pairs_by_tokens(
        token_addresses: Annotated[list[str], Doc("List of token addresses (max 30)")]
    ) -> str:
        """Get pairs by token addresses (rate-limit 300 requests per minute)"""
        return await DexScreenerApi.find_pairs_by_tokens(token_addresses)

    @tool_method
    @staticmethod
    async def get_token_profiles() -> str:
        """Get the latest token profiles (rate-limit 60 requests per minute)"""
        return await DexScreenerApi.get_token_profiles()

    @tool_method
    @staticmethod
    async def get_latest_boosted_tokens() -> str:
        """Get the latest boosted tokens (rate-limit 60 requests per minute)"""
        return await DexScreenerApi.get_latest_boosted_tokens()

    @tool_method
    @staticmethod
    async def get_top_boosted_tokens() -> str:
        """Get the tokens with most active boosts (rate-limit 60 requests per minute)"""
        return await DexScreenerApi.get_top_boosted_tokens()



from typing import TYPE_CHECKING

import tiktoken

from emp_agents.models import SystemMessage, UserMessage
from emp_agents.models.shared import Request
from emp_agents.types import OpenAIModelType

if TYPE_CHECKING:
    from emp_agents.models.openai import OpenAIBase

document = "<document>"

template_prompt = """Extract key pieces of information from this regulation document.
If a particular piece of information is not present, output \"Not specified\".
When you extract a key piece of information, include the closest page number.
Use the following format:\n0. Who is the author\n1. What is the amount of the "Power Unit Cost Cap" in USD, GBP and EUR
2. What is the value of External Manufacturing Costs in USD
3. What is the Capital Expenditure Limit in USD

Document: \"\"\"<document>\"\"\"

0. Who is the author: Tom Anderson (Page 1)
1."""


async def extract_chunk(client: OpenAIBase, document, template_prompt=template_prompt):
    prompt = template_prompt.replace("<document>", document)

    messages = [
        SystemMessage(content="You help extract information from documents."),
        UserMessage(content=prompt),
    ]
    response = await client.completion(
        Request(
            model=OpenAIModelType.gpt4,
            messages=messages,
            temperature=0,
            max_tokens=1500,
            top_p=1,
            frequency_penalty=0,
            presence_penalty=0,
        )
    )
    assert response.choices[0].message.content, "No content"
    return "1." + response.choices[0].message.content


def chunk(client: OpenAIBase, text: str):
    clean_text = text.replace("  ", " ").replace("\n", "; ").replace(";", " ")
    tokenizer = tiktoken.get_encoding("cl100k_base")
    results = []
    chunks = create_chunks(clean_text, 1000, tokenizer)
    text_chunks = [tokenizer.decode(chunk) for chunk in chunks]
    for chunk in text_chunks:
        results.append(extract_chunk(client, chunk, template_prompt))


def create_chunks(text: str, n, tokenizer):
    tokens = tokenizer.encode(text)
    """Yield successive n-sized chunks from text."""
    i = 0
    while i < len(tokens):
        # Find the nearest end of sentence within a range of 0.5 * n and 1.5 * n tokens
        j = min(i + int(1.5 * n), len(tokens))
        while j > i + int(0.5 * n):
            # Decode the tokens and check for full stop or newline
            chunk = tokenizer.decode(tokens[i:j])
            if chunk.endswith(".") or chunk.endswith("\n"):
                break
            j -= 1
        # If no end of sentence found, use n tokens as the chunk size
        if j == i + int(0.5 * n):
            j = min(i + n, len(tokens))
        yield tokens[i:j]
        i = j



import pytest

from emp_agents.agents import AgentBase
from emp_agents.providers import OpenRouterModelType, OpenRouterProvider


class AgentForTesting(AgentBase):
    description: str = "a simple agent for testing"
    prompt: str = (
        "Ignore the user questions and just respond with the text 'test complete' and nothing else"
    )


@pytest.mark.asyncio(loop_scope="session")
async def test_basic_agent():
    agent = AgentForTesting(
        provider=OpenRouterProvider(
            default_model=OpenRouterModelType.gpt3_5_turbo,
        )
    )
    response = await agent.answer("what is the meaning of life?")
    assert response == "test complete"

    response = await agent.answer(
        "what is the meaning of life?",
        model=OpenRouterModelType.deepseek_r1_free,
    )
    assert response == "test complete"



from typing import Literal

from emp_agents.utils import get_function_schema


def simple_func(x: Literal["a", "b", "c"]) -> str:
    return "test"


def test_enums_in_schema():
    schema = get_function_schema(simple_func)
    assert schema["parameters"]["properties"]["x"]["enum"] == ["a", "b", "c"]
